#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import json
import numpy
from .. enums import *
from .. events import *
from .. properties import *
from . fonts import FONTS
from . frame import Frame
from . path import Path


class Canvas(PropertySet):
    """
    Abstract base class for drawing canvas. Each derived class must provide
    specific implementations for several basic drawing methods to unify the
    drawing API across multiple backends.
    
    The canvas 'width' and 'height' is always set in original device units (e.g.
    pixels), however, all the drawing methods are using logical units, which
    includes current drawing scale and drawing region aka 'viewport'. This makes
    all the drawing resolution-independent and in case of drawing the same
    graphics multiple times with different resolution, it is sufficient to
    change the 'draw_scale', 'line_scale' and 'font_scale' properties
    accordingly. None of the properties of the graphics needs to be changed.
    
    Properties:
        
        width: int or float
            Specifies the available width of the canvas in device units.
        
        height: int or float
            Specifies the available height of the canvas in device units.
        
        draw_scale: int or float
            Specifies the scaling factor for drawing.
        
        line_scale: int or float
            Specifies the scaling factor for line thickness.
        
        line_color: pero.Color, (int,), str, None or UNDEF
            Specifies the line color as an RGB or RGBA tuple, hex code, name or
            pero.Color.
        
        line_alpha: int, None or UNDEF
            Specifies the line alpha channel as a value between 0 and 255, where
            0 is fully transparent and 255 fully opaque. If this value is set,
            it will overwrite the alpha channel of the final line color.
        
        line_width: int, float or UNDEF
            Specifies the line width.
        
        line_style: pero.LINE_STYLE or UNDEF
            Specifies the line drawing style as any item from the
            pero.LINE_STYLE enum.
        
        line_dash: (float,), None or UNDEF
            Specifies the line dash style as a collection of numbers defining the
            lengths of lines and spaces in-between. Specified value is used only
            if the 'line_style' property is set to pero.LINE_STYLE.CUSTOM.
        
        line_cap: pero.LINE_CAP or UNDEF
            Specifies the line ends shape as any item from the pero.LINE_CAP
            enum.
        
        line_join: pero.LINE_JOIN or UNDEF
            Specifies the line corners shape as any item from the
            pero.LINE_JOIN enum.
        
        fill_color: pero.Color, (int,), str, None or UNDEF
            Specifies the fill color as an RGB or RGBA tuple, hex code, name or
            pero.Color.
        
        fill_alpha: int, None or UNDEF
            Specifies the fill alpha channel as a value between 0 and 255,
            where 0 is fully transparent and 255 fully opaque. If this value is
            set, it will overwrite the alpha channel of the final fill color.
        
        fill_style: pero.FILL_STYLE or UNDEF
            Specifies the fill style as any item from the pero.FILL_STYLE
            enum.
        
        font_scale: int, float
            Specifies the scaling factor for texts.
        
        font_size: int, None or UNDEF
            Specifies the font size or None to reset to default size.
        
        font_name: str, None or UNDEF
            Specifies an existing font name or None to reset to default family.
        
        font_family: pero.FONT_FAMILY, None or UNDEF
            Specifies the font family as any item from the pero.FONT_FAMILY
            enum or None to reset to default family.
        
        font_style: pero.FONT_STYLE, None or UNDEF
            Specifies the font style as any item from the pero.FONT_STYLE
            enum or None to reset to default style.
        
        font_weight: pero.FONT_WEIGHT, None or UNDEF
            Specifies the font weight as any item from the pero.FONT_WEIGHT
            enum or None to reset to default weight.
        
        text_align: pero.TEXT_ALIGN, None or UNDEF
            Specifies the text alignment as any item from the pero.TEXT_ALIGN
            enum or None to reset to default alignment.
        
        text_base: pero.TEXT_BASE, None or UNDEF
            Specifies the text baseline as any item from the
            pero.TEXT_BASE enum or None to reset to default baseline.
        
        text_color: pero.Color, (int,), str, None or UNDEF
            Specifies the text foreground color as an RGB or RGBA tuple, hex
            code, name or pero.Color.
        
        text_alpha: int, None or UNDEF
            Specifies the text foreground alpha channel as a value between 0 and
            255, where 0 is fully transparent and 255 fully opaque. If this
            value is set, it will overwrite the alpha channel of the final text
            color.
        
        text_bgr_color: pero.Color, (int,), str, None or UNDEF
            Specifies the text background color as an RGB or RGBA tuple, hex
            code, name or pero.Color.
        
        text_bgr_alpha: int, None or UNDEF
            Specifies the text background alpha channel as a value between 0 and
            255, where 0 is fully transparent and 255 fully opaque. If this
            value is set, it will overwrite the alpha channel of the final text
            background color.
        
        text_split: bool
            Specifies whether the text should be first split into individual
            lines. This requires corresponding 'text_splitter' property to be
            set.
        
        text_splitter: str
            Specifies the character(s) to be used for splitting a text into
            individual lines.
        
        text_spacing: float
            Specifies additional space to be inserted between text lines as
            multiplier of line height.
        
        viewport: pero.Frame (read-only)
            Specifies current drawing region coordinates in logical units. It
            defines actual drawing origin and logical width and height of the
            canvas.
    """
    
    width = NumProperty(100, dynamic=False)
    height = NumProperty(100, dynamic=False)
    
    draw_scale = FloatProperty(1, dynamic=False)
    line_scale = FloatProperty(1, dynamic=False)
    font_scale = FloatProperty(1, dynamic=False)
    
    pen = Include(LineProperties, dynamic=False,
        line_color = "#000",
        line_width = 1,
        line_dash = None,
        line_style = LINE_STYLE_SOLID,
        line_cap = LINE_CAP_ROUND,
        line_join = LINE_JOIN_ROUND)
    
    brush = Include(FillProperties, dynamic=False,
        fill_color = None,
        fill_style = FILL_STYLE_SOLID)
    
    text = Include(TextProperties, dynamic=False,
        font_size = 10,
        font_family = FONT_FAMILY_SANS,
        font_style = FONT_STYLE_NORMAL,
        font_weight = FONT_WEIGHT_NORMAL,
        text_align = TEXT_ALIGN_LEFT,
        text_base = TEXT_BASE_TOP,
        text_color = "#000",
        text_bgr_color = None,
        text_split = True,
        text_splitter = LINE_SPLITTER,
        text_spacing = 0)
    
    
    def __init__(self, **overrides):
        """Initializes a new instance of Canvas."""
        
        super().__init__(**overrides)
        
        # hold properties
        for prop in LineProperties.properties():
            self.hold_property(prop.name)
        
        for prop in FillProperties.properties():
            self.hold_property(prop.name)
        
        for prop in TextProperties.properties():
            self.hold_property(prop.name)
        
        # init property names
        self._pen_properties = set(p.name for p in LineProperties.properties())
        self._brush_properties = set(p.name for p in FillProperties.properties())
        self._text_properties = set(p.name for p in TextProperties.properties())
        
        # init scale and offset
        self._scale = self.draw_scale
        self._offset = numpy.array((0, 0))
        
        # set font
        self._cfont = None
        
        # init views
        self._viewport = None
        self._viewport_full = Frame(0, 0, self.width/self.draw_scale, self.height/self.draw_scale)
        
        # bind events
        self.bind(EVT_PROPERTY_CHANGED, self._on_canvas_property_changed)
    
    
    @property
    def viewport(self):
        """
        Gets current drawing area coordinates.
        
        Returns:
            pero.Frame
                Current drawing area.
        """
        
        # get full view
        if self._viewport is None:
            return self._viewport_full
        
        # get current view
        return self._viewport
    
    
    def set_pen_by(self, prop_set, prefix="", source=UNDEF, overrides=None):
        """
        Extracts and applies all line properties from given property set.
        
        Args:
            prop_set: pero.PropertySet
                Object from which to extract line properties.
            
            prefix: str
                Prefix used for line properties to be extracted.
            
            source: ? or None
                Data source to be used for calculating callable properties.
            
            overrides:
                Specific properties to be overwritten.
        """
        
        # extract and set line properties
        for prop in self._pen_properties:
            value = prop_set.get_property(prefix+prop, source, overrides)
            self.set_property(prop, value, True)
    
    
    def set_brush_by(self, prop_set, prefix="", source=UNDEF, overrides=None):
        """
        Extracts and applies all fill properties from given property set.
        
        Args:
            prop_set: pero.PropertySet
                Object from which to extract fill properties.
            
            prefix: str
                Prefix used for fill properties to be extracted.
            
            source: ? or None
                Data source to be used for calculating callable properties.
            
            overrides:
                Specific properties to be overwritten.
        """
        
        # extract and set fill properties
        for prop in self._brush_properties:
            value = prop_set.get_property(prefix+prop, source, overrides)
            self.set_property(prop, value, True)
    
    
    def set_text_by(self, prop_set, prefix="", source=UNDEF, overrides=None):
        """
        Extracts and applies all text properties from given property set.
        
        Args:
            prop_set: pero.PropertySet
                Object from which to extract text properties.
            
            prefix: str
                Prefix used for text properties to be extracted.
            
            source: ? or None
                Data source to be used for calculating callable properties.
            
            overrides:
                Specific properties to be overwritten.
        """
        
        # extract and set text properties
        for prop in self._text_properties:
            value = prop_set.get_property(prefix+prop, source, overrides)
            self.set_property(prop, value, True)
    
    
    def get_font(self):
        """
        Gets current font.
        
        Returns:
            pero.Font
                Current font.
        """
        
        # get current font
        if self._cfont is not None:
            return self._cfont
        
        # get font names
        names = []
        if self.font_name:
            names.append(self.font_name)
        
        if self.font_family in FONT_FAMILY_NAMES:
            names += FONT_FAMILY_NAMES[self.font_family]
        
        # get specific font
        for name in names:
            
            self._cfont = FONTS.get_font(
                family = name,
                style = self.font_style,
                weight = self.font_weight,
                loose = True)
            
            if self._cfont is not None:
                return self._cfont
        
        # font not found
        message = "Cannot initialize the font! -> %s" % ", ".join(names)
        raise ValueError(message)
    
    
    def get_line_size(self, text):
        """
        Gets width and height of a single text line using current text settings.
        
        The result is given in logical units, i.e. current font scaling is
        applied but no line scaling. Note that if used directly to draw any line
        drawing (e.g. bounding rectangle), current line scaling will be applied.
        Be sure to apply inverse scaling before or use the 'get_text_size'
        method instead.
        
        This method should be overridden by specific backend to get more
        precise size.
        
        Args:
            text: str
                Text for which the size should be calculated.
        
        Returns:
            (float, float)
                Line width and height.
        """
        
        # check text
        if not text:
            return 0, 0
        
        # get font
        font = self.get_font()
        
        # get font size
        size = 11
        if self.font_size:
            size = int(0.5 + self.font_size * self.font_scale)
        
        # get text size
        width, height = font.get_size(text, size)
        
        # get full height
        full = font.get_size("j", size)
        height = max(height, full[1])
        
        return width, height
    
    
    def get_text_size(self, text, invert_scaling=True):
        """
        Gets width and height of the text bounding box using current text
        settings.
        
        By default, resulting width and height are inverse-scaled by current
        line scaling factor. It allows direct use for drawing text-related
        graphics like bounding rectangles etc. This behavior can be disabled
        by setting 'invert_scaling' to False.
        
        Args:
            text: str
                Text for which the size should be calculated.
            
            invert_scaling: bool
                If set to True, resulting width and height are inverse-scaled by
                current line scale factor.
        
        Returns:
            (float, float)
                Text width and height.
        """
        
        width = 0
        height = 0
        
        # check text
        if not text:
            return width, height
        
        # split lines
        lines = [text]
        if self.text_split and self.text_splitter:
            lines = text.split(self.text_splitter)
        
        # calc size
        for line in lines:
            
            # get line size
            line_width, line_height = self.get_line_size(line or " ")
            
            # update size
            height += line_height
            if line_width > width:
                width = line_width
        
        # add line spacing
        height += line_height*self.text_spacing * (len(lines) - 1)
        
        # invert scaling
        if invert_scaling:
            width /= self._scale
            height /= self._scale
        
        return width, height
    
    
    def get_text_bbox(self, text, x=0, y=0, angle=0, invert_scaling=True):
        """
        Gets bounding box of the text.
        
        By default, text width and height are inverse-scaled by current line
        scaling factor. It allows direct use for drawing text-related graphics
        like bounding rectangles etc. This behavior can be disabled by setting
        'invert_scaling' to False.
        
        Args:
            text: str
                Text to be drawn.
            
            x: int or float
                X-coordinate of the text anchor.
            
            y: int or float
                Y-coordinate of the text anchor.
            
            angle: float
                Text angle in radians.
            
            invert_scaling: bool
                If set to True, text width and height are inverse-scaled by
                current line scale factor.
        
        Returns:
            pero.Frame
                Text bounding box.
        """
        
        # check data
        if not text:
            return Frame(x, y, 0, 0)
        
        # init buffers
        boxes = []
        x0 = x
        y0 = y
        height = 0
        
        # split lines
        lines = [text]
        if self.text_split and self.text_splitter:
            lines = text.split(self.text_splitter)
        
        # calc boxes
        for line in lines:
            
            # get line size
            width, height = self.get_line_size(line or " ")
            
            # invert scaling
            if invert_scaling:
                width /= self._scale
                height /= self._scale
            
            # apply alignment
            if self.text_align == TEXT_ALIGN_CENTER:
                x0 = x - 0.5 * width
            
            elif self.text_align == TEXT_ALIGN_RIGHT:
                x0 = x - width
            
            # store box
            boxes.append(Frame(x0, y0, width, height))
            
            # add line spacing
            y0 += height * (1 + self.text_spacing)
        
        # apply baseline
        y_offset = 0
        total_height = y0 - y - height * self.text_spacing
        
        if self.text_base == TEXT_BASE_MIDDLE:
            y_offset = - 0.5*total_height
        
        elif self.text_base == TEXT_BASE_BOTTOM:
            y_offset = - total_height
        
        if y_offset:
            for box in boxes:
                box.offset(y=y_offset)
        
        # calc sin
        sin = numpy.sin(angle)
        cos = numpy.cos(angle)
        
        # make final frame
        frame = Frame(x, y)
        for box in boxes:
            for p in box.points:
                px = x + (p[0]-x) * cos - (p[1]-y) * sin
                py = y + (p[0]-x) * sin + (p[1]-y) * cos
                frame.extend(px, py)
        
        return frame
    
    
    def draw_arc(self, x, y, radius, start_angle, end_angle, clockwise=True):
        """
        Draws an arc of specified radius centered around given coordinates.
        
        This method should be overridden by specific backend to provide native
        implementation other than the default, using pero.Path.
        
        Args:
            x: int or float
                X-coordinate of the center.
            
            y: int or float
                Y-coordinate of the center.
            
            radius: int or float
                Radius of the arc.
            
            start_angle: int or float
                Angle of the starting point in radians.
            
            end_angle: int or float
                Angle of the starting point in radians.
            
            clockwise: bool
                Direction of drawing between start and end point.
        """
        
        # make path
        path = Path()
        path.arc(x, y, radius, start_angle, end_angle, clockwise)
        
        # draw path
        self.draw_path(path)
    
    
    def draw_bow(self, x1, y1, x2, y2, radius, large=False, clockwise=True):
        """
        Draws an arc specified by given radius and end-point coordinates. One
        of the four existing solutions is chosen according to the 'large' and
        'clockwise' parameters.
        
        Args:
            x1: int, float, callable
                Specifies the x-coordinate of the arc start.
            
            y1: int, float, callable
                Specifies the y-coordinate of the arc start.
            
            x2: int, float, callable
                Specifies the x-coordinate of the arc end.
            
            y2: int, float, callable
                Specifies the y-coordinate of the arc end.
            
            radius: int, float, callable
                Specifies the arc radius.
            
            large: bool
                Specifies which of the possible arcs will be drawn according
                to its length.
            
            clockwise: bool, callable
                Specifies which of the possible arcs will be drawn according to
                drawing direction. If set to True the clockwise arc is drawn,
                otherwise the anti-clockwise.
        """
        
        # get points distance and angle
        dist = 0.5 * numpy.sqrt((x2 - x1) ** 2 + (y2 - y1) ** 2)
        angle = numpy.arctan2(y2 - y1, x2 - x1)
        
        # check min radius
        radius = max(radius, dist)
        
        # calc sin/cos
        sin = numpy.sin(angle)
        cos = numpy.cos(angle)
        
        # get origins
        c = numpy.sqrt(radius ** 2 - dist ** 2)
        c1x = x1 + dist * cos - c * sin
        c1y = y1 + dist * sin + c * cos
        c2x = x1 + dist * cos + c * sin
        c2y = y1 + dist * sin - c * cos
        
        # set circles
        if angle >= 0:
            small = c1x, c1y
            big = c2x, c2y
        else:
            big = c2x, c2y
            small = c1x, c1y
        
        # apply direction
        if not clockwise:
            small, big = big, small
        
        # select final circle
        cx, cy = big if large else small
        
        # get angle
        start_angle = numpy.arctan2(y1 - cy, x1 - cx)
        end_angle = numpy.arctan2(y2 - cy, x2 - cx)
        
        # draw arc
        self.draw_arc(cx, cy, radius, start_angle, end_angle, clockwise)
    
    
    def draw_circle(self, x, y, radius):
        """
        Draws a circle of specified radius centered around given coordinates.
        
        This method should be overridden by specific backend to provide native
        implementation other than the default using, pero.Path.
        
        Args:
            x: int or float
                X-coordinate of the center.
            
            y: int or float
                Y-coordinate of the center.
            
            radius: int or float
                Radius of the circle.
        """
        
        # make path
        path = Path()
        path.circle(x, y, radius)
        
        # draw path
        self.draw_path(path)
    
    
    def draw_ellipse(self, x, y, width, height):
        """
        Draws an ellipse centered around given coordinates and fitting into the
        width and height.
        
        This method should be overridden by specific backend to provide native
        implementation other than the default, using pero.Path.
        
        Args:
            x: int or float
                X-coordinate of the center.
            
            y: int or float
                Y-coordinate of the center.
            
            width: int or float
                Full width of the ellipse.
            
            height: int or float
                Full height of the ellipse.
        """
        
        # make path
        path = Path()
        path.ellipse(x, y, width, height)
        
        # draw path
        self.draw_path(path)
    
    
    def draw_graphics(self, graphics, source=UNDEF, **overrides):
        """
        Draws given graphics into current canvas.
        
        Args:
            graphics: pero.Graphics
                Graphics to be drawn.
            
            source: any
                Data source to be used for calculating callable properties of
                the graphics.
            
            overrides: str:any pairs
                Specific properties of the graphics to be overwritten.
        """
        
        graphics.draw(self, source=source, **overrides)
    
    
    def draw_json(self, dump):
        """
        Draws graphics by given JSON dump. Such dump can be easily created by
        drawing to pero.Image and calling its 'get_json' method.
        
        Args:
            dump: str or dict
                JSON image dump.
        """
        
        # load json
        if not isinstance(dump, dict):
            dump = json.loads(dump)
        
        # call commands
        for name, args in dump.get("commands", []):
            
            # check UNDEF
            if name == 'set_property' and args['value'] == UNDEF:
                args = args.copy()
                args['value'] = UNDEF
            
            # recreate path
            if 'path' in args:
                args = args.copy()
                args['path'] = Path.from_json(args['path'])
            
            # call method
            method = getattr(self, name)
            method(**args)
    
    
    def draw_line(self, x1, y1, x2, y2):
        """
        Draws a line between two points.
        
        This method should be overridden by specific backend to provide native
        implementation other than the default, using pero.Path.
        
        Args:
            x1: int or float
                X-coordinate of the line start.
            
            y1: int or float
                Y-coordinate of the line start.
            
            x2: int or float
                X-coordinate of the line end.
            
            y2: int or float
                Y-coordinate of the line end.
        """
        
        # make path
        path = Path()
        path.move_to(x1, y1)
        path.line_to(x2, y2)
        
        # draw path
        self.draw_path(path)
    
    
    def draw_lines(self, points):
        """
        Draws continuous open line using sequence of points.
        
        This method should be overridden by specific backend to provide native
        implementation other than the default, using pero.Path.
        
        Args:
            points: ((float, float),)
                Sequence of x,y coordinates of the points.
        """
        
        # init path
        path = Path()
        path.move_to(points[0][0], points[0][1])
        
        # add points
        for p in points[1:]:
            path.line_to(p[0], p[1])
        
        # draw path
        self.draw_path(path)
    
    
    def draw_path(self, path):
        """
        Draws given path using current pen and brush.
        
        This method must be overwritten by specific backend to provide native
        implementation for path drawing.
        
        Args:
            path: pero.Path
                Path to be drawn.
        """
        
        raise NotImplementedError("The 'draw_path' method is not implemented for '%s'." % self.__class__.__name__)
    
    
    def draw_polygon(self, points):
        """
        Draws a closed polygon using sequence of points.
        
        This method should be overridden by specific backend to provide native
        implementation other than the default, using pero.Path.
        
        Args:
            points: ((float, float),)
                Sequence of x,y coordinates of the points.
        """
        
        # check points
        if len(points) < 2:
            return
        
        # init path
        path = Path()
        path.move_to(points[0][0], points[0][1])
        
        # add points
        for p in points[1:]:
            path.line_to(p[0], p[1])
        
        # close path
        path.close()
        
        # draw path
        self.draw_path(path)
    
    
    def draw_ray(self, x, y, angle, length, offset=0):
        """
        Draws a line defined by origin, angle and length.
        
        Args:
            x: int or float
                X-coordinate of the line start.
            
            y: int or float
                Y-coordinate of the line start.
            
            angle: int or float
                Line angle in radians.
            
            length: int or float
                Specifies the line length.
            
            offset: int or float
                Specifies the shift from the origin while keeping the length and
                angle.
        """
        
        # calc end point
        if angle:
            x = x + offset * numpy.cos(angle)
            y = y + offset * numpy.sin(angle)
            x2 = x + length * numpy.cos(angle)
            y2 = y + length * numpy.sin(angle)
        else:
            x = x + offset
            x2 = x + length
            y2 = y
        
        # draw path
        self.draw_line(x, y, x2, y2)
    
    
    def draw_rect(self, x, y, width, height, radius=None):
        """
        Draws a rectangle specified by given top left corner and size and
        optional round corners specified as a single value or individual value
        for each corners starting from top-left.
        
        This method should be overridden by specific backend to provide native
        implementation other than the default, using pero.Path.
        
        Args:
            x: int or float
                X-coordinate of the top left corner.
            
            y: int or float
                Y-coordinate of the top left corner.
            
            width: int or float
                Full width of the rectangle.
            
            height: int or float
                Full height of the rectangle.
            
            radius: int, float, (int,int,int,int) or (float,float,float,float)
                Radius of curved corners.
        """
        
        # make path
        path = Path()
        path.rect(x, y, width, height, radius)
        
        # draw path
        self.draw_path(path)
    
    
    def draw_text(self, text, x, y, angle=0):
        """
        Draws a text string anchored at specified point using current text
        settings.
        
        This method must be overwritten by specific backend to provide native
        implementation for text drawing.
        
        Args:
            text: str
                Text to be drawn.
            
            x: int or float
                X-coordinate of the text anchor.
            
            y: int or float
                Y-coordinate of the text anchor.
            
            angle: int or float
                Text angle in radians.
        """
        
        raise NotImplementedError("The 'draw_text' method is not implemented for '%s'." % self.__class__.__name__)
    
    
    def draw_text_polar(self, text, x, y, radius, angle, position=POS_CENTER, rotation=TEXT_ROT_FOLLOW):
        """
        Draws a text string anchored at specified point with automatic alignment
        and baseline according to given angle, using current text settings.
        
        Args:
            text: str
                Text to be drawn.
            
            x: int or float
                X-coordinate of the text anchor.
            
            y: int or float
                Y-coordinate of the text anchor.
            
            radius: int or float
                Radius of the virtual anchor circle.
            
            angle: int or float
                Text anchor angle in radians.
            
            position: pero.POSITION_IOC
                Text position along the anchor circle as any item from the
                pero.POSITION_IOC enum.
            
            rotation: pero.TEXT_ROTATION
                Relative rotation of the text along the anchor circle as any
                item from the pero.TEXT_ROTATION enum.
        """
        
        # remember current settings
        ori_text_align = self.text_align
        ori_text_base = self.text_base
        
        # set angle tolerance
        tol = radius * numpy.sin(numpy.deg2rad(4))
        
        # rotate coordinates
        x1 = x + radius * numpy.cos(angle)
        y1 = y + radius * numpy.sin(angle)
        
        # get position
        is_perpendicular = (rotation in (TEXT_ROT_NATURAL, TEXT_ROT_FACEOUT, TEXT_ROT_FACEIN))
        is_out = (position == POS_OUTSIDE)
        
        # adjust angle
        if rotation == TEXT_ROT_FACEOUT:
            angle += 0.5*numpy.pi
        
        elif rotation == TEXT_ROT_FACEIN:
            angle -= 0.5*numpy.pi
        
        elif rotation == TEXT_ROT_NATURAL and y1 > y + tol:
            angle -= 0.5*numpy.pi
        
        elif rotation == TEXT_ROT_NATURAL:
            angle += 0.5*numpy.pi
        
        elif x1 < x - tol:
            angle -= numpy.pi
        
        # perpendicular
        if is_perpendicular:
            self.text_align = TEXT_ALIGN_CENTER
        
        # on edge
        elif position == POS_CENTER:
            self.text_align = TEXT_ALIGN_CENTER
        
        # left side
        elif x1 < x - tol:
            self.text_align = TEXT_ALIGN_RIGHT if is_out else TEXT_ALIGN_LEFT
        
        # right side
        elif x1 > x + tol:
            self.text_align = TEXT_ALIGN_LEFT if is_out else TEXT_ALIGN_RIGHT
        
        # rotated at center
        elif rotation == TEXT_ROT_FOLLOW:
            self.text_align = TEXT_ALIGN_LEFT if is_out else TEXT_ALIGN_RIGHT
        
        # at center
        else:
            self.text_align = TEXT_ALIGN_CENTER
        
        # center on edge
        if position == POS_CENTER:
            self.text_base = TEXT_BASE_MIDDLE
        
        # rotated
        elif rotation == TEXT_ROT_FOLLOW:
            self.text_base = TEXT_BASE_MIDDLE
            
        # perpendicular
        elif rotation == TEXT_ROT_FACEOUT:
            self.text_base = TEXT_BASE_BOTTOM if is_out else TEXT_BASE_TOP
            
        # perpendicular
        elif rotation == TEXT_ROT_FACEIN:
            self.text_base = TEXT_BASE_TOP if is_out else TEXT_BASE_BOTTOM
            
        # top part
        elif y1 < y - tol:
            self.text_base = TEXT_BASE_BOTTOM if is_out else TEXT_BASE_TOP
        
        # bottom part
        elif y1 > y + tol:
            self.text_base = TEXT_BASE_TOP if is_out else TEXT_BASE_BOTTOM
        
        # rotated at middle
        elif rotation == TEXT_ROT_NATURAL:
            self.text_base = TEXT_BASE_BOTTOM if is_out else TEXT_BASE_TOP
        
        # at middle
        else:
            self.text_base = TEXT_BASE_MIDDLE
        
        # draw text
        if rotation != TEXT_ROT_NONE:
            self.draw_text(text, x1, y1, angle=angle)
        else:
            self.draw_text(text, x1, y1)
        
        # reset original settings
        self.text_align = ori_text_align
        self.text_base = ori_text_base
    
    
    def fill(self, color=UNDEF):
        """
        Fills current drawing region by specified or actual fill color.
        
        Args:
            color: pero.Color, (int,), str, None or UNDEF
                Specifies the fill color as an RGB or RGBA tuple, hex code, name
                or pero.Color. If not set, current fill color will be used.
        """
        
        # remove current pen
        line_width = self.line_width
        self.line_width = 0
        
        # set fill color
        fill_color = self.fill_color
        if color:
            self.fill_color = color
        
        # draw rectangle
        self.draw_rect(*self.viewport.rect)
        
        # reset pen and brush
        self.line_width = line_width
        self.fill_color = fill_color
    
    
    def view(self, x=None, y=None, width=None, height=None, relative=False):
        """
        Sets rectangular region currently used for drawing. This provides an
        easy way to draw complex graphics at specific position of the canvas
        without adjusting the coordinates of the graphics. It is achieved by
        changing the origin coordinates and the logical width and height of
        the canvas.
        
        This method returns a context manager class so it can be used as part of
        the 'with' statement to make the changes temporary and automatically
        reverted back to original state.
        
        Args:
            x: int or float
                X-coordinate of the top-left corner.
            
            y: int or float
                Y-coordinate of the top-left corner.
            
            width: int, float or None
                Full width of the viewport.
            
            height: int, float or None
                Full height of the viewport.
            
            relative: bool
                If set to True the new viewport is applied relative to current
                one.
        
        Returns:
            pero.ViewState
                Viewport state context manager.
        """
        
        # get current viewport
        viewport = self.viewport
        
        # init state
        state = ViewState(self, viewport)
        
        # reset to full
        if x is None and y is None:
            self._viewport = None
            self._offset = numpy.array((0, 0))
            return state
        
        # get origin
        if x is None:
            x = 0
        if y is None:
            y = 0
        
        # make relative
        if relative:
            x += viewport.x
            y += viewport.y
        
        # get width
        if width is None:
            width = viewport.width
        
        # get height
        if height is None:
            height = viewport.height
        
        # set viewport
        self._viewport = Frame(x, y, width, height)
        self._offset = numpy.array((x, y))
        
        # return state
        return state
    
    
    def clip(self, path):
        """
        Sets clipping path as intersection with current one.
        
        This method returns a context manager class so it can be used as part of
        the 'with' statement to make the changes temporary and automatically
        reverted back to original state.
        
        This method needs be overwritten by specific backend to provide native
        implementation for clipping.
        
        Args:
            path: pero.Path
                Path to be used for clipping.
        
        Returns:
            pero.ClipState
                Clipping state context manager.
        """
        
        # return state
        return ClipState(self)
    
    
    def unclip(self):
        """
        Removes last clipping path while keeping previous if any.
        
        This method needs be overwritten by specific backend to provide native
        implementation for clipping.
        """
        
        pass
    
    
    def group(self, id_tag=None, class_tag=None):
        """
        Opens new drawing group.
        
        This method returns a context manager class so it can be used as part of
        the 'with' statement to make the changes temporary and automatically
        reverted back to original state.
        
        This method needs be overwritten by specific backend to provide native
        implementation for objects grouping.
        
        Args:
            id_tag: str
                Unique id of the group.
            
            class_tag:
                Class of the group.
        
        Returns:
            pero.GroupState
                Grouping state context manager.
        """
        
        # return state
        return GroupState(self)
    
    
    def ungroup(self):
        """
        Closes the last drawing group.
        
        This method needs be overwritten by specific backend to provide native
        implementation for objects grouping.
        """
        
        pass
    
    
    def to_device(self, x, y):
        """
        Recalculates given logical content position into absolute device
        position within the canvas by removing current line scale and origin
        offset.
        
        Args:
            x: float
                Logical x-coordinate.
            
            y: float
                Logical y-coordinate.
        
        Returns:
            (float, float)
                Absolute device x and y coordinates.
        """
        
        x = (x + self._offset[0]) * self._scale
        y = (y + self._offset[1]) * self._scale
        
        return x, y
    
    
    def to_logical(self, x, y):
        """
        Recalculates given absolute device position into logical position within
        current view by applying current line scale and origin offset.
        
        Args:
            x: float
                Device x-coordinate.
            
            y: float
                Device y-coordinate.
        
        Returns:
            (float, float)
                Logical x and y coordinates within current view.
        """
        
        x = x / self._scale - self._offset[0]
        y = y / self._scale - self._offset[1]
        
        return x, y
    
    
    def _on_canvas_property_changed(self, evt):
        """Called after any property has changed."""
        
        # update global scaling
        if evt.name == 'draw_scale':
            self._scale = self.draw_scale
        
        # update full viewport
        if evt.name in ('draw_scale', 'width', 'height'):
            self._viewport_full = Frame(0, 0, self.width/self.draw_scale, self.height/self.draw_scale)
        
        # update current font
        if evt.name in ('font_name', 'font_family', 'font_style', 'font_weight'):
            self._cfont = None
        
        # update current pen
        if evt.name in self._pen_properties or evt.name == 'line_scale':
            self.fire(PenChangedEvt.from_evt(evt))
        
        # update current brush
        if evt.name in self._brush_properties:
            self.fire(BrushChangedEvt.from_evt(evt))
        
        # update current text
        if evt.name in self._text_properties or evt.name == 'font_scale':
            self.fire(TextChangedEvt.from_evt(evt))


class CanvasState(object):
    """
    Allows using 'with' statement for canvas state methods like 'view', 'clip'
    or 'group'.
    """
    
    
    def __init__(self, canvas, *args, **kwargs):
        """
        Initializes a new instance of CanvasState.
        
        Args:
            canvas: pero.Canvas
                Current canvas.
        """
        
        self._canvas = canvas
    
    
    def __enter__(self):
        """The 'with' statement started."""
        
        return self
    
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        """The 'with' statement ended."""
        
        pass


class ClipState(CanvasState):
    """Allows using 'with' statement for canvas 'clip' method."""
    
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        """Removes clipping when 'with' statement ended."""
        
        self._canvas.unclip()


class GroupState(CanvasState):
    """Allows using 'with' statement for canvas 'group' method."""
    
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        """Removes grouping when 'with' statement ended."""
        
        self._canvas.ungroup()


class ViewState(CanvasState):
    """Allows using 'with' statement for canvas 'view' method."""
    
    
    def __init__(self, canvas, viewport, *args, **kwargs):
        """
        Initializes a new instance of ViewState.
        
        Args:
            canvas: pero.Canvas
                Current canvas.
            
            viewport: pero.Frame or None
                Viewport state to keep.
        """
        
        super().__init__(canvas)
        self._viewport = viewport
    
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        """Resets original viewport when 'with' statement ended."""
        
        if self._viewport is None:
            self._canvas.view(None)
        else:
            self._canvas.view(*self._viewport.rect)



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import fitz
import numpy
from ... properties import *
from ... drawing import Canvas, Path, Matrix
from . enums import *


class MuPDFCanvas(Canvas):
    """Wrapper for PyMuPDF document page."""
    
    
    def __init__(self, page, **overrides):
        """
        Initializes a new instance of MuPDFCanvas.
        
        Args:
            page: fritz.Page
                PyMuPDF document page.
            
            overrides: str:any pairs
                Value overwrites for specific properties.
        """
        
        # init buffers
        self._page = page
        
        self._pen = {
            'width': None,
            'color': (0, 0, 0),
            'cap': MUPDF_LINE_CAP[SQUARE],
            'join': MUPDF_LINE_JOIN[BEVEL],
            'dashes': "[]0"}
        
        self._brush = {
            'color': None}
        
        self._font = {
            'name': 'helvetica',
            'file': None,
            'size': 10,
            'descent': 0,
            'for_color': (0, 0, 0),
            'bgr_color': None}
        
        # init size
        rect = self._page.bound()
        
        if 'width' not in overrides:
            overrides['width'] = rect.width
        
        if 'height' not in overrides:
            overrides['height'] = rect.height
        
        # init base
        super().__init__(**overrides)
        
        # init canvas
        self._update_pen()
        self._update_brush()
        self._update_text()
        
        # bind events
        self.bind(EVT_PEN_CHANGED, self._update_pen)
        self.bind(EVT_BRUSH_CHANGED, self._update_brush)
        self.bind(EVT_TEXT_CHANGED, self._update_text)
    
    
    def draw_circle(self, x, y, radius):
        """
        Draws a circle of specified radius centered around given coordinates.
        
        Args:
            x: int or float
                X-coordinate of the center.
            
            y: int or float
                Y-coordinate of the center.
            
            radius: int or float
                Radius of the circle.
        """
        
        # apply scaling and offset
        x = self._scale * (x + self._offset[0])
        y = self._scale * (y + self._offset[1])
        radius = self._scale * radius
        
        # init shape
        shape = self._page.new_shape()
        
        # draw
        shape.draw_circle((x, y), radius)
        
        # fill and stroke
        self._fill_and_stroke(shape, close=True)
    
    
    def draw_ellipse(self, x, y, width, height):
        """
        Adds an ellipse centered at given position and fitting into the size.
        
        Args:
            x: int or float
                X-coordinate of the center.
            
            y: int or float
                Y-coordinate of the center.
            
            width: int or float
                Full width of the ellipse.
            
            height: int or float
                Full height of the ellipse.
        """
        
        # apply scaling and offset
        x = self._scale * (x + self._offset[0])
        y = self._scale * (y + self._offset[1])
        width = self._scale * width
        height = self._scale * height
        
        # check size
        if not width or not height:
            return
        
        # init shape
        shape = self._page.new_shape()
        
        # draw
        shape.draw_oval((x-0.5*width, y-0.5*height, x+0.5*width, y+0.5*height))
        
        # fill and stroke
        self._fill_and_stroke(shape, close=True)
    
    
    def draw_line(self, x1, y1, x2, y2):
        """
        Draws a line between two points.
        
        Args:
            x1: int or float
                X-coordinate of the line start.
            
            y1: int or float
                Y-coordinate of the line start.
            
            x2: int or float
                X-coordinate of the line end.
            
            y2: int or float
                Y-coordinate of the line end.
        """
        
        # apply scaling and offset
        x1 = self._scale * (x1 + self._offset[0])
        y1 = self._scale * (y1 + self._offset[1])
        x2 = self._scale * (x2 + self._offset[0])
        y2 = self._scale * (y2 + self._offset[1])
        
        # init shape
        shape = self._page.new_shape()
        
        # draw
        shape.draw_line((x1, y1), (x2, y2))
        
        # fill and stroke
        self._fill_and_stroke(shape, close=False, fill=False)
    
    
    def draw_lines(self, points):
        """
        Draws continuous open line using sequence of points.
        
        Args:
            points: ((float, float),)
                Sequence of x,y coordinates of the points.
        """
        
        # check points
        if len(points) < 2:
            return
        
        # apply scaling and offset
        points = (numpy.array(points) + self._offset) * numpy.array((self._scale, self._scale))
        
        # init shape
        shape = self._page.new_shape()
        
        # draw
        shape.draw_polyline(points)
        
        # fill and stroke
        self._fill_and_stroke(shape, close=False)
    
    
    def draw_path(self, path):
        """
        Draws the path using current pen and brush.
        
        Args:
            path: pero.Path
                Path to be drawn.
        """
        
        # apply scaling and offset
        matrix = Matrix()
        matrix.translate(self._offset[0], self._offset[1])
        matrix.scale(self._scale, self._scale)
        path = path.transformed(matrix)
        
        # init shape
        shape = self._page.new_shape()
        cursor = (0, 0)
        origin = (0, 0)
        
        # draw
        for command in path.commands():
            
            # get data
            key = command[0]
            values = command[1:]
            
            # close
            if key == PATH_CLOSE:
                shape.draw_line(cursor, origin)
                cursor = origin
            
            # move to
            elif key == PATH_MOVE:
                cursor = values[0:2]
                origin = cursor
            
            # line to
            elif key == PATH_LINE:
                shape.draw_line(cursor, values[0:2])
                cursor = values[0:2]
            
            # curve to
            elif key == PATH_CURVE:
                shape.draw_bezier(cursor, values[0:2], values[2:4], values[4:6])
                cursor = values[4:6]
        
        # fill and stroke
        self._fill_and_stroke(shape, close=False, even_odd=MUPDF_FILL_RULE[path.fill_rule])
    
    
    def draw_polygon(self, points):
        """
        Draws a closed polygon using sequence of points.
        
        Args:
            points: ((float, float),)
                Sequence of x,y coordinates of the points.
        """
        
        # check points
        if len(points) < 2:
            return
        
        # apply scaling and offset
        points = (numpy.array(points) + self._offset) * numpy.array((self._scale, self._scale))
        
        # init shape
        shape = self._page.new_shape()
        
        # draw
        shape.draw_polyline(points)
        
        # fill and stroke
        self._fill_and_stroke(shape, close=True)
    
    
    def draw_rect(self, x, y, width, height, radius=None):
        """
        Draws a rectangle specified by given top left corner and size and
        optional round corners specified as a single value or individual value
        for each corner starting from top-left.
        
        Args:
            x: int or float
                X-coordinate of the top left corner.
            
            y: int or float
                Y-coordinate of the top left corner.
            
            width: int or float
                Full width of the rectangle.
            
            height: int or float
                Full height of the rectangle.
            
            radius: int, float, (int,int,int,int) or (float,float,float,float)
                Radius of curved corners.
        """
        
        # get radius
        if isinstance(radius, (int, float)) and radius != 0:
            radius = (radius, radius, radius, radius)
        
        # draw as path if having round corners
        if radius:
            path = Path()
            path.rect(x, y, width, height, radius)
            self.draw_path(path)
            return
        
        # apply scaling and offset
        x = self._scale * (x + self._offset[0])
        y = self._scale * (y + self._offset[1])
        width = self._scale * width
        height = self._scale * height
        
        # init shape
        shape = self._page.new_shape()
        
        # draw
        shape.draw_rect((x, y, x+width, y+height))
        
        # fill and stroke
        self._fill_and_stroke(shape, close=True)
    
    
    def draw_text(self, text, x, y, angle=0):
        """
        Draws a text string anchored at specified point using current text
        settings.
        
        Args:
            text: str
                Text to be drawn.
            
            x: int or float
                X-coordinate of the text anchor.
            
            y: int or float
                Y-coordinate of the text anchor.
            
            angle: int or float
                Text angle in radians.
        """
        
        # init shape
        shape = self._page.new_shape()
        
        # get full size
        full_width, full_height = self.get_text_size(text)
        descent = self._font['descent'] / self._scale
        
        # get angle transformation
        text_morph = None
        bgr_morph = None
        
        if angle:
            angle *= -1
            
            matrix = fitz.Matrix(numpy.rad2deg(angle))
            text_morph = [None, matrix]
            bgr_morph = [None, matrix]
            
            sin = numpy.sin(angle)
            cos = numpy.cos(angle)
        
        # split lines
        lines = [text]
        if self.text_split and self.text_splitter:
            lines = text.split(self.text_splitter)
        
        # draw lines
        for i, line in enumerate(lines):
            
            # get line size
            line_width, line_height = self.get_line_size(line)
            line_width /= self._scale
            line_height /= self._scale
            
            # init offset
            x_offset = 0
            y_offset = line_height - descent
            
            # adjust alignment
            if self.text_align == TEXT_ALIGN_CENTER:
                x_offset -= 0.5 * line_width
            
            elif self.text_align == TEXT_ALIGN_RIGHT:
                x_offset -= line_width
            
            # adjust baseline
            if self.text_base == TEXT_BASE_MIDDLE:
                y_offset -= 0.5*full_height
            
            elif self.text_base == TEXT_BASE_BOTTOM:
                y_offset -= full_height
            
            # add line offset
            y_offset += i * line_height * (1 + self.text_spacing)
            
            # init coords
            text_x = x_offset
            text_y = y_offset
            bgr_x = x_offset
            bgr_y = y_offset + descent
            
            # apply angle
            if angle:
                
                x_shift = text_x * cos + text_y * sin
                y_shift = -text_x * sin + text_y * cos
                text_x = x_shift
                text_y = y_shift
                
                x_shift = bgr_x * cos + bgr_y * sin
                y_shift = -bgr_x * sin + bgr_y * cos
                bgr_x = x_shift
                bgr_y = y_shift
            
            # apply scaling and offset
            text_x = self._scale * (x + text_x + self._offset[0])
            text_y = self._scale * (y + text_y + self._offset[1])
            
            bgr_x1 = self._scale * (x + bgr_x + self._offset[0])
            bgr_y2 = self._scale * (y + bgr_y + self._offset[1])
            bgr_x2 = bgr_x1 + line_width * self._scale
            bgr_y1 = bgr_y2 - line_height * self._scale
            
            # set angle pivot
            if angle:
                text_morph[0] = fitz.Point(text_x, text_y)
                bgr_morph[0] = fitz.Point(bgr_x1, bgr_y2)
            
            # draw background
            if self._font['bgr_color']:
                
                shape = self._page.new_shape()
                shape.draw_rect((bgr_x1, bgr_y1, bgr_x2, bgr_y2))
                
                shape.finish(
                    width = 0,
                    color = None,
                    fill = self._font['bgr_color'],
                    morph = bgr_morph,
                    closePath = True)
            
            # draw text
            shape.insert_text(
                (text_x, text_y),
                line,
                fontsize = self._font['size'],
                fontname = self._font['name'],
                fontfile = self._font['file'],
                set_simple = False,
                encoding = 0,
                color = None,
                fill = self._font['for_color'],
                render_mode = 0,
                morph = text_morph)
            
            # update page
            shape.commit()
    
    
    def _fill_and_stroke(self, shape, close, even_odd=True, fill=True, commit=True):
        """Fills and strokes given shape."""
        
        # draw shape
        shape.finish(
            width = self._pen['width'] if self._pen['color'] else 0,
            color = self._pen['color'] if self._pen['width'] else None,
            lineCap = self._pen['cap'],
            lineJoin = self._pen['join'],
            dashes = self._pen['dashes'],
            fill = self._brush['color'] if fill else None,
            even_odd = even_odd,
            closePath = close)
        
        # update page
        if commit:
            shape.commit()
    
    
    def _update_pen(self, evt=None):
        """Updates pen with current properties."""
        
        # get property name
        prop_name = evt.name if evt is not None else None
        
        # update color
        if prop_name is None or prop_name in ('line_color', 'line_alpha'):
            color = ColorProperties.get_color(self, "line_")
            if color is not UNDEF:
                self._pen['color'] = color.rgb_r if color.alpha else None
        
        # update width
        if prop_name is None or prop_name in ('line_width', 'line_scale'):
            line_width = self.line_width
            if line_width is not UNDEF:
                self._pen['width'] = line_width * self.line_scale
        
        # update cap
        if prop_name is None or prop_name == 'line_cap':
            line_cap = self.line_cap
            if line_cap is not UNDEF:
                self._pen['cap'] = MUPDF_LINE_CAP[line_cap]
        
        # update join
        if prop_name is None or prop_name == 'line_join':
            line_join = self.line_join
            if line_join is not UNDEF:
                self._pen['join'] = MUPDF_LINE_JOIN[line_join]
        
        # update style/dash
        if prop_name is None or prop_name in ('line_dash', 'line_style', 'line_width', 'line_scale'):
            line_style = self.line_style
            line_dash = self.line_dash if self.line_dash else []
            line_width = self._pen['width']
            
            if line_style == LINE_STYLE_SOLID:
                line_dash = []
            elif line_style not in (LINE_STYLE_CUSTOM, UNDEF):
                line_dash = MUPDF_LINE_STYLE[line_style]
            
            self._pen['dashes'] = "[%s]0" % " ".join(str(int(x*line_width)) for x in line_dash)
    
    
    def _update_brush(self, evt=None):
        """Updates brush with current properties."""
        
        color = ColorProperties.get_color(self, "fill_")
        
        if self.fill_style == FILL_STYLE_TRANS:
            self._brush['color'] = None
        
        elif color is not UNDEF:
            self._brush['color'] = color.rgb_r if color.alpha else None
    
    
    def _update_text(self, evt=None):
        """Updates text with current properties."""
        
        # get property name
        prop_name = evt.name if evt is not None else None
        
        # update font
        if prop_name is None or prop_name in ('font_name', 'font_family', 'font_style', 'font_weight'):
            font = self.get_font()
            name = "%s %s %s" % (font.name, font.style, font.weight)
            self._font['name'] = name.title().replace(" ", "")
            self._font['file'] = font.path
        
        # update font size
        if prop_name is None or prop_name in ('font_size', 'font_scale'):
            font_size = self.font_size
            if font_size is None:
                self._font['size'] = 11 * self.font_scale
            elif font_size is not UNDEF:
                self._font['size'] = font_size * self.font_scale
        
        # update font descent
        if prop_name is None or prop_name in ('font_name', 'font_family', 'font_style', 'font_weight', 'font_size', 'font_scale'):
            font = self.get_font()
            self._font['descent'] = font.get_descent(int(0.5+self._font['size']))
        
        # update foreground color
        if prop_name is None or prop_name in ('text_color', 'text_alpha'):
            color = ColorProperties.get_color(self, "text_")
            if color is not UNDEF:
                self._font['for_color'] = color.rgb_r if color.alpha else None
        
        # update background color
        if prop_name is None or prop_name in ('text_bgr_color', 'text_bgr_alpha'):
            color = ColorProperties.get_color(self, "text_bgr_")
            if color is not UNDEF:
                self._font['bgr_color'] = color.rgb_r if color.alpha else None



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import unittest
import pero


class TestCase(unittest.TestCase):
    """Test case for Gradient class."""
    
    
    def test_constructor(self):
        """Tests whether constructor works correctly."""
        
        # full specification
        colors = (pero.Color.Red, pero.Color.Green, pero.Color.Blue)
        stops = (0, 0.5, 1)
        gradient = pero.Gradient(colors, stops, "Name")
        
        self.assertEqual(gradient.colors, colors)
        self.assertEqual(gradient.stops, stops)
        self.assertEqual(gradient.name, "Name")
        
        # no name
        colors = (pero.Color.Red, pero.Color.Green, pero.Color.Blue)
        stops = (0, 0.5, 1)
        gradient = pero.Gradient(colors, stops)
        
        self.assertEqual(gradient.colors, colors)
        self.assertEqual(gradient.stops, stops)
        self.assertEqual(gradient.name, None)
        
        # no stops
        colors = (pero.Color.Red, pero.Color.Green, pero.Color.Blue)
        stops = (0, 0.5, 1)
        gradient = pero.Gradient(colors)
        
        self.assertEqual(gradient.colors, colors)
        self.assertEqual(gradient.stops, stops)
        self.assertEqual(gradient.name, None)
        
        # stops range
        colors = (pero.Color.Red, pero.Color.Green, pero.Color.Blue)
        stops = (100, 150, 200)
        gradient = pero.Gradient(colors, (100, 200))
        
        self.assertEqual(gradient.colors, colors)
        self.assertEqual(gradient.stops, stops)
        self.assertEqual(gradient.name, None)
        
        # test invalid color
        colors = (pero.Color.Red, pero.Color.Green, "blue", 0)
        
        with self.assertRaises(ValueError):
            gradient = pero.Gradient(colors)
        
        # test invalid stop
        colors = (pero.Color.Red, pero.Color.Green, pero.Color.Blue)
        stops = (0, 0.5, "1")
        
        with self.assertRaises(TypeError):
            gradient = pero.Gradient(colors, stops)
        
        # test unsorted stops
        colors = (pero.Color.Red, pero.Color.Green, pero.Color.Blue)
        stops = (0, 1, 0.5)
        
        with self.assertRaises(ValueError):
            gradient = pero.Gradient(colors, stops)
        
        # test unequal size
        colors = (pero.Color.Red, pero.Color.Green, pero.Color.Blue)
        stops = (0, 0.25, 0.5, 0.75, 1)
        
        with self.assertRaises(ValueError):
            gradient = pero.Gradient(colors, stops)
    
    
    def test_color_at(self):
        """Tests whether gradient color is generated correctly."""
        
        # init gradient
        colors = (pero.Color.Red, pero.Color.Green, pero.Color.Blue)
        stops = (0, 0.5, 1)
        gradient = pero.Gradient(colors, stops, "Name")
        
        # test below range
        color = gradient.color_at(-1)
        self.assertEqual(color, colors[0])
        
        # test above range
        color = gradient.color_at(2)
        self.assertEqual(color, colors[-1])
        
        # test first
        color = gradient.color_at(0)
        self.assertEqual(color, colors[0])
        
        # test middle
        color = gradient.color_at(0.5)
        self.assertEqual(color, colors[1])
        
        # test last
        color = gradient.color_at(1)
        self.assertEqual(color, colors[-1])
        
        # test interpolated
        color = gradient.color_at(0.25)
        model = pero.Color(128,64,0,255)
        self.assertEqual(color, model)
    
    
    def test_normalized(self):
        """Tests whether gradient normalization works correctly."""
        
        # init gradient
        colors = (pero.Color.Red, pero.Color.Green, pero.Color.Blue)
        stops = (10, 60, 110)
        gradient = pero.Gradient(colors, stops)
        
        # test normalized
        normalized = gradient.normalized()
        model = (0, .5, 1)
        self.assertEqual(normalized.stops, model)
        
        normalized = gradient.normalized(0, 1)
        model = (0, .5, 1)
        self.assertEqual(normalized.stops, model)
        
        normalized = gradient.normalized(100, 1100)
        model = (100, 600, 1100)
        self.assertEqual(normalized.stops, model)
    
    
    def test_from_name(self):
        """Tests whether named gradients can be accessed."""
        
        # init named gradient
        colors = (pero.Color.Red, pero.Color.Green, pero.Color.Blue)
        model = pero.Gradient(colors, name="MyGradient")
        
        # test from name
        gradient = pero.Gradient.from_name('MyGradient')
        gradient = pero.Gradient.from_name('mygradient')
        
        # test from palette name
        gradient = pero.Gradient.from_name('Spectral')
        gradient = pero.Gradient.from_name('spectral')
    
    
    def test_name_registering(self):
        """Tests whether new named gradients are registered."""
        
        # init named gradient
        colors = (pero.Color.Red, pero.Color.Green, pero.Color.Blue)
        model = pero.Gradient(colors, name="MyGradient")
        
        # test from class
        gradient = pero.Gradient.MyGradient
        self.assertTrue(gradient is model)
        
        gradient = pero.Gradient.mygradient
        self.assertTrue(gradient is model)
        
        # test from lib
        gradient = pero.GRADIENTS.MyGradient
        self.assertTrue(gradient is model)
        
        gradient = pero.GRADIENTS.mygradient
        self.assertTrue(gradient is model)


# run test case
if __name__ == "__main__":
    unittest.main(verbosity=2)



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

from .. enums import *
from . event import Event


class ViewEvt(Event):
    """
    Abstract base class for various types of view events.
    
    Attributes:
        
        native: any
            Native event fired by the view.
            
        view: pero.View
            The view, which fires the event.
            
        control: pero.Control
            The control which fired the event.
    """
    
    TYPE = EVT_VIEW
    
    
    def __init__(self, **kwargs):
        """Initializes a new instance of ViewEvt."""
        
        self.native = None
        self.view = None
        self.control = None
        
        super().__init__(**kwargs)
    
    
    @classmethod
    def from_evt(cls, evt):
        """
        Initializes a new instance of given class by copying all data.
        
        Args:
            evt: pero.ViewEvt
                Source event from which to copy the data.
        
        Returns:
            cls instance
                New instance of requested class.
        """
        
        return cls(
            native = evt.native,
            view = evt.view,
            control = evt.control)


class SizeEvt(ViewEvt):
    """
    Defines an event which is fired if view size was changed.
    
    Attributes:
        
        width: int or float
            New width of the view.
        
        height: int or float
            New height of the view.
    """
    
    TYPE = EVT_SIZE
    
    
    def __init__(self, **kwargs):
        """Initializes a new instance of SizeEvt."""
        
        self.width = None
        self.height = None
        
        super().__init__(**kwargs)


class ZoomEvt(ViewEvt):
    """Defines an event which is fired if axes ranges were changed."""
    
    TYPE = EVT_ZOOM



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

# import enums
from . values import *
from . maths import *
from . splits import *
from . drawing import *
from . positions import *
from . text import *
from . backends import *
from . views import *
from . events import *
from . keys import *



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

from .. enums import *
from .. properties import *
from . glyph import Glyph


class Grid(Glyph):
    """
    Abstract base class for various types of grids, typically used to draw plot
    grids. They are drawn as series of lines for given coordinates defined by
    the 'ticks' property and specified origin. Every tick value represents a
    single coordinate according to a grid type etc. The origin coordinates
    represent a top-left corner of the bounding box or a center in centered
    grids.
    
    Grid serve as a drawing glyph only and do not implement any logic to
    generate the ticks itself. Everything is expected to be supplied in the
    final form already.
    
    Properties:
        
        x: int, float or callable
            Specifies the x-coordinate of the origin.
        
        y: int, float or callable
            Specifies the y-coordinate of the origin.
        
        ticks: (float,), callable, None or UNDEF
            Specifies the sequence of ticks values according to particular
            type of grid.
        
        line properties:
            Includes pero.LineProperties to specify the line.
    """
    
    x = NumProperty(0)
    y = NumProperty(0)
    ticks = TupleProperty(None, intypes=(int, float), nullable=True)
    
    pen = Include(LineProperties)


class ParallelGrid(Grid):
    """
    Parallel grid is drawn as a series of horizontal or vertical parallel
    lines of specified length. This type of grid is typically used by
    Cartesian plots.
    
    According to the 'relative' property, the ticks are expected to be provided
    as relative values (True) (as distances from the origin) or as absolute
    values (False) (as distances from device zero). The coordination of the
    distances is defined by the 'orientation' property.
    
    Properties:
        
        orientation: pero.ORIENTATION or callable
            Specifies the lines orientation as any item from the
            pero.ORIENTATION enum.
        
        relative: bool or callable
            Specifier whether the ticks values are given as a shift from the
            origin (True) or as absolute values (False).
        
        length: int, float or callable
            Specifies the lines length.
        
        angle properties:
            Includes pero.AngleProperties to specify the lines angle.
    """
    
    orientation = EnumProperty(ORI_HORIZONTAL, enum=ORIENTATION)
    relative = BoolProperty(False)
    length = NumProperty(UNDEF)
    angle = Include(AngleProperties)
    
    
    def draw(self, canvas, source=UNDEF, **overrides):
        """Uses given canvas to draw grid."""
        
        # check if visible
        if not self.is_visible(source, overrides):
            return
        
        # get properties
        tag = self.get_property('tag', source, overrides)
        ticks = self.get_property('ticks', source, overrides)
        x = self.get_property('x', source, overrides)
        y = self.get_property('y', source, overrides)
        orientation = self.get_property('orientation', source, overrides)
        relative = self.get_property('relative', source, overrides)
        length = self.get_property('length', source, overrides)
        angle = AngleProperties.get_angle(self, '', ANGLE_RAD, source, overrides)
        
        # check ticks
        if not ticks:
            return
        
        # set pen and brush
        canvas.set_pen_by(self, source=source, overrides=overrides)
        canvas.fill_color = None
        
        # apply orientation
        if orientation == ORI_VERTICAL:
            angle -= 0.5*math.pi
            length *= -1
        
        # make ticks relative
        offset = 0
        if not relative:
            offset -= x if orientation == ORI_VERTICAL else y
        
        # calc sin/cos
        sin = round(math.sin(angle), 5)
        cos = round(math.cos(angle), 5)
        
        # start drawing group
        canvas.group(tag, "grid")
        
        # draw lines
        for tick in ticks:
            
            x1 = x - (tick+offset) * sin
            y1 = y + (tick+offset) * cos
            x2 = x + length * cos - (tick+offset) * sin
            y2 = y + length * sin + (tick+offset) * cos
            
            canvas.draw_line(x1=x1, y1=y1, x2=x2, y2=y2)
        
        # end drawing group
        canvas.ungroup()


class RayGrid(Grid):
    """
    Ray grid is drawn as a star-like series of lines of specified length.
    This type of grid is typically used by polar plots.
    
    The 'ticks' are expected to be provided in angle 'units'.
    
    Properties:
        
        length: int, float or callable
            Specifies the lines length.
        
        offset: int, float or callable
            Specifies the shift from center to be applied to each tick.
        
        units: pero.ANGLE or callable
            Specifies the angle units for the ticks as any item from the
            pero.ANGLE enum.
    """
    
    length = NumProperty(UNDEF)
    offset = NumProperty(0)
    units = EnumProperty(ANGLE_RAD, enum=ANGLE)
    
    
    def draw(self, canvas, source=UNDEF, **overrides):
        """Uses given canvas to draw grid."""
        
        # check if visible
        if not self.is_visible(source, overrides):
            return
        
        # get properties
        tag = self.get_property('tag', source, overrides)
        ticks = self.get_property('ticks', source, overrides)
        units = self.get_property('units', source, overrides)
        x = self.get_property('x', source, overrides)
        y = self.get_property('y', source, overrides)
        offset = self.get_property('offset', source, overrides)
        length = self.get_property('length', source, overrides)
        
        # check ticks
        if not ticks:
            return
        
        # set pen and brush
        canvas.set_pen_by(self, source=source, overrides=overrides)
        canvas.fill_color = None
        
        # convert angles
        if units == ANGLE_DEG:
            ticks = tuple(map(math.radians, ticks))
        
        # get radii
        inner_radius = offset
        outer_radius = offset + length
        
        # start drawing group
        canvas.group(tag, "grid")
        
        # draw lines
        for angle in ticks:
            
            cos = math.cos(angle)
            sin = math.sin(angle)
            
            x1 = x + inner_radius * cos
            y1 = y + inner_radius * sin
            x2 = x + outer_radius * cos
            y2 = y + outer_radius * sin
            
            canvas.draw_line(x1, y1, x2, y2)
        
        # end drawing group
        canvas.ungroup()


class RadialGrid(Grid):
    """
    Radial grid is drawn as series of concentric circles or arcs defined
    by 'start_angle' and 'end_angle'. This type of grid is typically used
    by polar plots.
    
    The 'ticks' are expected to be provided as distance values from the origin.
    
    Properties:
        
        start_angle properties:
            Includes pero.AngleProperties to specify the start angle.
        
        end_angle properties:
            Includes pero.AngleProperties to specify the end angle.
        
        clockwise: bool or callable
            Specifies the drawing direction. If set to True the grid is drawn
            clockwise, otherwise anti-clockwise.
    """
    
    start_angle = Include(AngleProperties, prefix="start_")
    end_angle = Include(AngleProperties, prefix="end_")
    clockwise = BoolProperty(True)
    
    
    def draw(self, canvas, source=UNDEF, **overrides):
        """Uses given canvas to draw grid."""
        
        # check if visible
        if not self.is_visible(source, overrides):
            return
        
        # get properties
        tag = self.get_property('tag', source, overrides)
        ticks = self.get_property('ticks', source, overrides)
        x = self.get_property('x', source, overrides)
        y = self.get_property('y', source, overrides)
        clockwise = self.get_property('clockwise', source, overrides)
        start_angle = AngleProperties.get_angle(self, 'start_', ANGLE_RAD, source, overrides)
        end_angle = AngleProperties.get_angle(self, 'end_', ANGLE_RAD, source, overrides)
        
        # check ticks
        if not ticks:
            return
        
        # set pen and brush
        canvas.set_pen_by(self, source=source, overrides=overrides)
        canvas.fill_color = None
        
        # start drawing group
        canvas.group(tag, "grid")
        
        # draw full circles
        if start_angle % (2 * math.pi) == end_angle % (2 * math.pi):
            for radius in ticks:
                canvas.draw_circle(x, y, radius)
        
        # draw arcs
        else:
            for radius in ticks:
                canvas.draw_arc(x, y, radius, start_angle, end_angle, clockwise)
        
        # end drawing group
        canvas.ungroup()



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import os.path
from . loader import QSizeF, QPageSize, QImage, QPrinter, QPainter, QApplication
from . enums import *
from . canvas import QtCanvas
from . viewer import QtViewer


def show(graphics, title=None, width=None, height=None, **options):
    """
    Shows given graphics in the viewer app.
    
    Args:
        graphics: pero.Graphics or pero.Control
            Graphics to be shown.
        
        title: str or None
            Viewer frame title.
        
        width: float or None
            Viewer width in device units.
        
        height: float or None
            Viewer height in device units.
    """
    
    # init app
    app = QApplication.instance()
    if app is None:
        app = QApplication([])
    
    # init main window
    window = QtViewer()
    
    # set title
    if title is not None:
        window.set_title(title)
    
    # check size
    if not width:
        width = VIEWER_WIDTH
    if not height:
        height = VIEWER_HEIGHT
    
    # set size
    window.set_size((width, height))
    
    # set graphics
    window.set_content(graphics)
    
    # draw graphics
    window.refresh()
    
    # start app
    window.show()
    app.exec()


def export(graphics, path, width=None, height=None, **options):
    """
    Draws given graphics into specified image file. The image format is
    determined from the extension of given file path. Supported extensions are
    .bmp, .gif, .jpg, .jpeg, .pdf and .png.
    
    Args:
        graphics: pero.Graphics
            Graphics to be drawn.
        
        path: str
            Full path of a file to save the image into.
        
        width: float or None
            Image width in device units.
        
        height: float or None
            Image height in device units.
        
        draw_scale: float
            Drawing scaling factor.
        
        line_scale: float
            Line scaling factor.
        
        font_scale: float
            Font scaling factor.
        
        quality: int
            Image quality in range between 0 and 100 with 0 meaning very poor
            and 100 excellent. This option is only available for JPEG format.
    """
    
    # get filename and extension
    dirname, filename = os.path.split(path)
    basename, extension = os.path.splitext(filename)
    extension = extension.lower()
    
    # export as raster image
    if extension in QT_RASTER_TYPES:
        export_raster(graphics, path, width, height, **options)
    
    # export as vector format
    elif extension in QT_VECTOR_TYPES:
        export_vector(graphics, path, width, height, **options)
    
    # unsupported image format
    else:
        message = "Unsupported image format! -> %s" % extension
        raise NotImplementedError(message)


def export_raster(graphics, path, width=None, height=None, **options):
    """
    Draws given graphics as raster image into specified file. The image format
    is determined from the extension of given file path. Supported extensions
    are .bmp, .gif, .jpg, .jpeg and .png.
    
    Args:
        graphics: pero.Graphics
            Graphics to be drawn.
        
        path: str
            Full path of a file to save the image into.
        
        width: float or None
            Image width in device units.
        
        height: float or None
            Image height in device units.
        
        draw_scale: float
            Drawing scaling factor.
        
        line_scale: float
            Line scaling factor.
        
        font_scale: float
            Font scaling factor.
        
        quality: int
            Image quality in range between 0 and 100 with 0 meaning very poor
            and 100 excellent.
    """
    
    # check size
    if not width:
        width = EXPORT_WIDTH
    if not height:
        height = EXPORT_HEIGHT
    
    # init app
    app = QApplication.instance()
    if app is None:
        app = QApplication([])
    
    # init image
    image = QImage(width, height, QImage.Format.Format_ARGB32)
    
    # init painter
    qp = QPainter()
    qp.begin(image)
    qp.setRenderHint(QPainter.RenderHint.Antialiasing)
    qp.setRenderHint(QPainter.RenderHint.SmoothPixmapTransform)
    
    # init canvas
    canvas = QtCanvas(qp, width=width, height=height)
    
    if 'draw_scale' in options:
        canvas.draw_scale = options['draw_scale']
    
    if 'line_scale' in options:
        canvas.line_scale = options['line_scale']
    
    if 'font_scale' in options:
        canvas.font_scale = options['font_scale']
    
    # draw graphics
    graphics.draw(canvas)
    
    # end drawing
    qp.end()
    
    # get image options
    quality = options.get('quality', -1)
    
    # save to file
    image.save(path, quality=quality)


def export_vector(graphics, path, width=None, height=None, **options):
    """
    Draws given graphics as vector image into specified file. The image format
    is determined from the extension of given file path. Supported extension
    is .pdf only.
    
    Args:
        graphics: pero.Graphics
            Graphics to be drawn.
        
        path: str
            Full path of a file to save the image into.
        
        width: float or None
            Image width in device units.
        
        height: float or None
            Image height in device units.
        
        draw_scale: float
            Drawing scaling factor.
        
        line_scale: float
            Line scaling factor.
        
        font_scale: float
            Font scaling factor.
    """
    
    # check size
    if not width:
        width = EXPORT_WIDTH
    if not height:
        height = EXPORT_HEIGHT
    
    # init app
    app = QApplication.instance()
    if app is None:
        app = QApplication([])
    
    # init page size
    size = QPageSize(QSizeF(width, height), QPageSize.Unit.Point)
    
    # init printer
    printer = QPrinter()
    printer.setPageSize(size)
    printer.setOutputFormat(QPrinter.OutputFormat.PdfFormat)
    printer.setOutputFileName(path)
    
    # init painter
    qp = QPainter()
    qp.begin(printer)
    qp.setRenderHint(QPainter.RenderHint.Antialiasing)
    qp.setRenderHint(QPainter.RenderHint.SmoothPixmapTransform)
    
    # init canvas
    canvas = QtCanvas(qp, width=width, height=height)
    
    if 'draw_scale' in options:
        canvas.draw_scale = options['draw_scale']
    
    if 'line_scale' in options:
        canvas.line_scale = options['line_scale']
    
    if 'font_scale' in options:
        canvas.font_scale = options['font_scale']
    
    # draw graphics
    graphics.draw(canvas)
    
    # end drawing
    qp.end()



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import unittest
import pero


class TestCase(unittest.TestCase):
    """Test case for stringed formatter."""
    
    
    def test_empty_formatter(self):
        """Tests whether formatter works as standalone tool."""
        
        formatter = pero.EmptyFormatter()
        
        self.assertEqual(formatter.format(0), "")
        self.assertEqual(formatter.format(1), "")
        self.assertEqual(formatter.format(1e6), "")
    
    
    def test_str_formatter(self):
        """Tests whether formatter works as standalone tool."""
        
        formatter = pero.StrFormatter(template="{:.2f} u")
        
        self.assertEqual(formatter.format(0), "0.00 u")
        self.assertEqual(formatter.format(1.5), "1.50 u")
        self.assertEqual(formatter.format(1e6), "1000000.00 u")
    
    
    def test_str_formatter_trim(self):
        """Tests whether formatter work with auto trimming."""
        
        formatter = pero.StrFormatter(template="\n{}")
        
        formatter.trim = False
        self.assertEqual(formatter.format("test"), "\ntest")
        
        formatter.trim = True
        self.assertEqual(formatter.format("test"), "test")
    
    
    def test_printf_formatter(self):
        """Tests whether formatter works as standalone tool."""
        
        formatter = pero.PrintfFormatter(template="%0.2f u")
        
        self.assertEqual(formatter.format(0), "0.00 u")
        self.assertEqual(formatter.format(1.5), "1.50 u")
        self.assertEqual(formatter.format(1e6), "1000000.00 u")
    
    
    def test_printf_formatter_trim(self):
        """Tests whether formatter work with auto trimming."""
        
        formatter = pero.PrintfFormatter(template="\n%s")
        
        formatter.trim = False
        self.assertEqual(formatter.format("test"), "\ntest")
        
        formatter.trim = True
        self.assertEqual(formatter.format("test"), "test")


# run test case
if __name__ == "__main__":
    unittest.main(verbosity=2)



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

from . values import *
from . enum import Enum

# define line styles
LINE_STYLE_CUSTOM = CUSTOM
LINE_STYLE_SOLID = SOLID
LINE_STYLE_DOTTED = DOTTED
LINE_STYLE_DASHED = DASHED
LINE_STYLE_DASHDOTTED = DASHDOTTED

LINE_STYLE = Enum(
    CUSTOM = LINE_STYLE_CUSTOM,
    SOLID = LINE_STYLE_SOLID,
    DOTTED = LINE_STYLE_DOTTED,
    DASHED = LINE_STYLE_DASHED,
    DASHDOTTED = LINE_STYLE_DASHDOTTED)

# define line caps
LINE_CAP_BUTT = BUTT
LINE_CAP_SQUARE = SQUARE
LINE_CAP_ROUND = ROUND

LINE_CAP = Enum(
    BUTT = LINE_CAP_BUTT,
    SQUARE = LINE_CAP_SQUARE,
    ROUND = LINE_CAP_ROUND)

# define line joins
LINE_JOIN_BEVEL = BEVEL
LINE_JOIN_MITER = MITER
LINE_JOIN_ROUND = ROUND

LINE_JOIN = Enum(
    BEVEL = LINE_JOIN_BEVEL,
    MITER = LINE_JOIN_MITER,
    ROUND = LINE_JOIN_ROUND)

# define line step
LINE_STEP_NONE = NONE
LINE_STEP_BEFORE = BEFORE
LINE_STEP_MIDDLE = MIDDLE
LINE_STEP_AFTER = AFTER

LINE_STEP = Enum(
    NONE = LINE_STEP_NONE,
    BEFORE = LINE_STEP_BEFORE,
    MIDDLE = LINE_STEP_MIDDLE,
    AFTER = LINE_STEP_AFTER)

# define line dash values
DASH_VALUES_DOTTED = (1, 5)
DASH_VALUES_DASHED = (5, 5)
DASH_VALUES_DASHDOTTED = (5, 5, 1, 5)

DASH_VALUES = Enum(
    DOTTED = DASH_VALUES_DOTTED,
    DASHED = DASH_VALUES_DASHED,
    DASHDOTTED = DASH_VALUES_DASHDOTTED)

# define fill style
FILL_STYLE_SOLID = SOLID
FILL_STYLE_TRANS = TRANS

FILL_STYLE = Enum(
    SOLID = FILL_STYLE_SOLID,
    TRANS = FILL_STYLE_TRANS)

# define fill rule
FILL_RULE_EVENODD = EVENODD
FILL_RULE_WINDING = WINDING

FILL_RULE = Enum(
    EVENODD = FILL_RULE_EVENODD,
    WINDING = FILL_RULE_WINDING)

# define path commands
PATH_CLOSE = 'Z'
PATH_MOVE = 'M'
PATH_LINE = 'L'
PATH_CURVE = 'C'

PATH = Enum(
    CLOSE = PATH_CLOSE,
    MOVE = PATH_MOVE,
    LINE = PATH_LINE,
    CURVE = PATH_CURVE)

# define available markers
MARKER_ASTERISK = '*'
MARKER_CIRCLE = 'o'
MARKER_CROSS = 'x'
MARKER_PLUS = '+'
MARKER_TRIANGLE = 't'
MARKER_SQUARE = 's'
MARKER_DIAMOND = 'd'
MARKER_PENTAGON = 'p'
MARKER_HEXAGON = 'h'

MARKER = Enum(
    ASTERISK = MARKER_ASTERISK,
    CIRCLE = MARKER_CIRCLE,
    CROSS = MARKER_CROSS,
    PLUS = MARKER_PLUS,
    TRIANGLE = MARKER_TRIANGLE,
    SQUARE = MARKER_SQUARE,
    DIAMOND = MARKER_DIAMOND,
    PENTAGON = MARKER_PENTAGON,
    HEXAGON = MARKER_HEXAGON)

# define available arrows
ARROW_ARC = 'c'
ARROW_BOW = ')'
ARROW_CONNECT_LINE = 'z'
ARROW_CONNECT_CURVE = 's'
ARROW_CURVE = '~'
ARROW_LINE = '-'
ARROW_RAY = '/'

ARROW = Enum(
    ARC = ARROW_ARC,
    BOW = ARROW_BOW,
    CONNECT_LINE = ARROW_CONNECT_LINE,
    CONNECT_CURVE = ARROW_CONNECT_CURVE,
    CURVE = ARROW_CURVE,
    LINE = ARROW_LINE,
    RAY = ARROW_RAY)

# define available heads
HEAD_CIRCLE = 'o'
HEAD_LINE = '|'
HEAD_NORMAL = '|>'
HEAD_NORMAL_B = '<|'
HEAD_OPEN = '>'
HEAD_OPEN_B = '<'
HEAD_VEE = '>>'
HEAD_VEE_B = '<<'

HEAD = Enum(
    CIRCLE = HEAD_CIRCLE,
    LINE = HEAD_LINE,
    NORMAL = HEAD_NORMAL,
    NORMAL_B = HEAD_NORMAL_B,
    OPEN = HEAD_OPEN,
    OPEN_B = HEAD_OPEN_B,
    VEE = HEAD_VEE,
    VEE_B = HEAD_VEE_B)



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

# Adapted from the original code by Pomax
# https://pomax.github.io/bezierinfo/

# import modules
import numpy
from . import helpers
from . frame import Frame

# define constants
PROJECTION_STEP = 0.1
REDUCE_STEP = 0.1
LUT_STEPS = 100
INTERSECT_THRESHOLD = 0.1


class Bezier(object):
    """
    Represents a cubic Bezier curve. This is just an experimental object to be
    able to calculate various attributes of a cubic Bezier curve.
    """
    
    
    def __init__(self, x1, y1, cx1, cy1, cx2, cy2, x2, y2):
        """
        Initialize a new instance of cubic Bezier curve.
        
        Args:
            x1: int or float
                X-coordinate of the start point.
            
            y1: int or float
                Y-coordinate of the start point.
            
            cx1: int or float
                X-coordinate of the start control point.
            
            cy1: int or float
                Y-coordinate of the start control point.
            
            cx2: int or float
                X-coordinate of the end control point.
            
            cy2: int or float
                Y-coordinate of the end control point.
            
            x2: int or float
                X-coordinate of the end point.
            
            y2: int or float
                Y-coordinate of the end point.
        """
        
        self._x1 = float(x1)
        self._y1 = float(y1)
        self._p1 = (self._x1, self._y1)
        
        self._cx1 = float(cx1)
        self._cy1 = float(cy1)
        self._c1 = (self._cx1, self._cy1)
        
        self._cx2 = float(cx2)
        self._cy2 = float(cy2)
        self._c2 = (self._cx2, self._cy2)
        
        self._x2 = float(x2)
        self._y2 = float(y2)
        self._p2 = (self._x2, self._y2)
        
        self._t1 = 0.
        self._t2 = 1.
        self._lut = []
        
        self._simple = None
        self._bbox = None
        self._extremes = None
        self._inflections = None
        self._length = None
        
        self._derivs = helpers.derivatives(*self.points)
    
    
    def __str__(self):
        """Gets standard string representation."""
        
        return "C(%d,%d %d,%d %d,%d %d,%d)" % (
            self._x1, self._y1,
            self._cx1, self._cy1,
            self._cx2, self._cy2,
            self._x2, self._y2)
    
    
    def __repr__(self):
        """Gets debug string representation."""
        
        return "%s(%s)" % (self.__class__.__name__, self.__str__())
    
    
    @property
    def coords(self):
        """
        Gets curve coordinates as x1, y1, cx1, cy1, cx2, cy2, x2, y2.
        
        Returns:
            (float, float, float, float, float, float, float, float)
                Curve coordinates.
        """
        
        return (
            self._x1, self._y1,
            self._cx1, self._cy1,
            self._cx2, self._cy2,
            self._x2, self._y2)
    
    
    @property
    def points(self):
        """
        Gets curve coordinates as p1, c1, c2, p2.
        
        Returns:
            ((float, float), (float, float), (float, float), (float, float))
                Curve coordinates.
        """
        
        return self._p1, self._c1, self._c2, self._p2
    
    
    def simple(self):
        """
        Gets a value indicating if current curve is simple, where a simpleness
        is defined as having all control points on the same side of the
        baseline, the control-to-end-point lines may not cross and the angle
        between the end point normals is no greater than 60 degrees.
        
        Returns:
            bool
                Returns True if simple, False otherwise.
        """
        
        if self._simple is None:
            
            a1 = helpers.angle(self._p2, self._p1, self._c1)
            a2 = helpers.angle(self._p2, self._p1, self._c1)
            
            if (a1 > 0 > a2) or (a1 < 0 < a2):
                self._simple = False
            
            else:
                nx1, ny1 = self.normal(0)
                nx2, ny2 = self.normal(1)
                
                if None in (nx1, ny1, nx2, ny2):
                    self._simple = True
                
                else:
                    s = nx1*nx2 + ny1*ny2
                    a = abs(numpy.arccos(s))
                    self._simple = a < numpy.pi/3
        
        return self._simple
    
    
    def bbox(self):
        """
        Calculates current curve bounding box based on anchors and extremes.
        
        Returns:
            pero.Frame
                Curve bounding box.
        """
        
        if self._bbox is None:
            
            self._bbox = Frame(self._x1, self._y1, self._x2-self._x1, self._y2-self._y1)
            
            extremes = self.extremes()
            extremes = set(extremes[0]+extremes[1])
            
            for t in extremes:
                x, y = self.point(t)
                self._bbox.extend(x, y)
        
        return self._bbox.clone()
    
    
    def length(self):
        """
        Calculates the length of current curve using numerical approximation.
        
        Returns:
            float
                Current curve length.
        """
        
        if self._length is None:
            self._length = 0
            
            for i in range(len(helpers.T_VALUES)):
                t = 0.5 * helpers.T_VALUES[i] + 0.5
                dx, dy = self.derivative(t)
                self._length += helpers.C_VALUES[i] * numpy.sqrt(dx*dx + dy*dy)
        
            self._length *= .5
        
        return self._length
    
    
    def extremes(self):
        """
        Calculates all extremes of current curve. For each dimension it provides
        all the t-values at which the extremes occur.
        
        Returns:
            ((float,), (float,))
                Extremes t-values for each dimension.
        """
        
        if self._extremes is None:
            
            ex = [[], []]
            
            for dim in (0, 1):
                s = lambda v: v[dim]
                
                p = list(map(s, self._derivs[0]))
                ex[dim] = helpers.droots(p)
                
                p = list(map(s, self._derivs[1]))
                ex[dim] += helpers.droots(p)
                
                f = lambda v: (0 <= v <= 1)
                ex[dim] = list(filter(f, ex[dim]))
            
            self._extremes = tuple(ex)
        
        return self._extremes
    
    
    def inflections(self):
        """
        Calculates all the inflection points of current curve as their t-values.
        
        Returns:
            (float,)
                Inflections t-values.
        """
        
        if self._inflections is None:
            
            p = helpers.align(self._p1, self._p2, *self.points)
            
            a = float(p[2][0] * p[1][1])
            b = float(p[3][0] * p[1][1])
            c = float(p[1][0] * p[2][1])
            d = float(p[3][0] * p[2][1])
            
            v1 = 18 * (-3*a + 2*b + 3*c - d)
            v2 = 18 * (3*a - b - 3*c)
            v3 = 18 * (c - a)
            
            if helpers.equals(v1, 0, 1e-3):
                
                if not helpers.equals(v2, 0, 1e-3):
                    t = -v3 / v2
                    if 0 <= t <= 1:
                        return [t]
                return []
            
            d = 2*v1
            if helpers.equals(d, 0, 1e-3):
                return []
            
            trm = v2*v2 - 4*v1*v3
            if trm < 0:
                return []
            
            sq = numpy.sqrt(v2*v2 - 4*v1*v3)
            buff = []
            
            t1 = (sq-v2)/d
            if 0 <= t1 <= 1:
                buff.append(t1)
            
            t2 = -(v2+sq)/d
            if 0 <= t2 <= 1:
                buff.append(t2)
            
            self._inflections = tuple(buff)
        
        return self._inflections
    
    
    def lut(self, steps=LUT_STEPS):
        """
        Generates a look-up table of coordinates on current curve, spaced at
        parametrically equidistant intervals of 1/steps.
        
        Args:
            steps: int
                Number of steps to generate.
        
        Returns:
            ((float, float),)
                Positions look-up table.
        """
        
        if len(self._lut) == steps+1:
            return self._lut
        
        self._lut = []
        for t in range(steps+1):
            self._lut.append(self.point(float(t)/steps))
        
        self._lut = tuple(self._lut)
        
        return self._lut
    
    
    def point(self, t):
        """
        Calculates x and y coordinates of the curve point at specified t-value.
        Note that the t-values are not distributed linearly along the curve but
        are influenced by the control points.
        
        Args:
            t: float
                The t-value position in range of 0-1.
        
        Returns:
            (float, float)
                Coordinates of the curve point.
        """
        
        if t == 0:
            return self._p1
        if t == 1:
            return self._p2
        
        mt = 1 - t
        a = mt*mt*mt
        b = mt*mt*t*3
        c = mt*t*t*3
        d = t*t*t
        
        x = a*self._x1 + b*self._cx1 + c*self._cx2 + d*self._x2
        y = a*self._y1 + b*self._cy1 + c*self._cy2 + d*self._y2
        
        return x, y
    
    
    def derivative(self, t):
        """
        Calculates the curve tangent at the specified t-value as a not-
        normalized vector.
        
        Args:
            t: float
                The t-value position in range of 0-1.
        
        Returns:
            (float, float)
                Derivatives vector.
        """
        
        mt = 1 - t
        a = mt*mt
        b = mt*t*2
        c = t*t
        
        p = self._derivs[0]
        
        dx = a*p[0][0] + b*p[1][0] + c*p[2][0]
        dy = a*p[0][1] + b*p[1][1] + c*p[2][1]
        
        return dx, dy
    
    
    def tangent(self, t):
        """
        Calculates the curve tangent at the specified t-value as a normalized
        vector.
        
        Args:
            t: float
                The t-value position in range of 0-1.
        
        Returns:
            (float, float)
                Normalized tangent vector.
        """
        
        dx, dy = self.derivative(t)
        q = numpy.sqrt(dx*dx + dy*dy)
        
        if dx == 0 and dy == 0:
            return None, None
        
        return dx/q, dy/q
    
    
    def normal(self, t):
        """
        Calculates the curve normal at the specified t-value as a normalized
        vector.
        
        Args:
            t: float
                The t-value position in range of 0-1.
        
        Returns:
            (float, float)
                Normalized normal vector.
        """
        
        dx, dy = self.derivative(t)
        
        if dx == 0 and dy == 0:
            print("Bad curve:", self)
            return None, None
        
        q = numpy.sqrt(dx*dx + dy*dy)
        
        return -dy/q, dx/q
    
    
    def hull(self, t):
        """
        Calculates the hull points for all iterations at specified t-value.
        This generates in total 10 points grouped by iterations. The first
        iteration contains 4 points, the second iteration contains 3 points, the
        third iteration contains 2 points and finally the fourth iteration
        contains just a single point, which is the point of the curve.
        
        Args:
            t: float
                The t-value position in range of 0-1.
        
        Returns:
            (((float, float),),)
                Coordinates of the hull points.
        """
        
        p = self.points
        h = [list(p)]
        
        while len(p) > 1:
            
            buff = []
            for i in range(len(p)-1):
                buff.append(helpers.relative(p[i], p[i+1], t))
            
            h.append(buff)
            p = buff
        
        return tuple(h)
    
    
    def split(self, t):
        """
        Splits current curve at specified t-value position. For the edge values
        of t-value the edge segment is set to None and a clone of current curve
        is returned for the other segment.
        
        Args:
            t: float
                The t-value position in range of 0-1.
        
        Returns:
            (pero.Bezier, pero.Bezier)
                Two new curves as a result of splitting.
        """
        
        if t == 0:
            return None, self.clone()
        if t == 1:
            return self.clone(), None
        
        h = self.hull(t)
        
        x1, y1 = h[0][0]
        cx1, cy1 = h[1][0]
        cx2, cy2 = h[2][0]
        x2, y2 = h[3][0]
        left = Bezier(x1, y1, cx1, cy1, cx2, cy2, x2, y2)
        
        x1, y1 = h[3][0]
        cx1, cy1 = h[2][1]
        cx2, cy2 = h[1][2]
        x2, y2 = h[0][3]
        right = Bezier(x1, y1, cx1, cy1, cx2, cy2, x2, y2)
        
        left._t1 = self._scale_t(0, 0, 1, self._t1, self._t2)
        left._t2 = self._scale_t(t, 0, 1, self._t1, self._t2)
        right._t1 = self._scale_t(t, 0, 1, self._t1, self._t2)
        right._t2 = self._scale_t(1, 0, 1, self._t1, self._t2)
        
        return left, right
    
    
    def slice(self, t1, t2):
        """
        Makes a slice of current curve at specified t-value positions.
        
        Args:
            t1: float
                The starting t-value position in range of 0-1.
            
            t2: float
                The end t-value position in range of 0-1.
        
        Returns:
            pero.Bezier
                Slice of current curve.
        """
        
        if t1 == t2:
            return None
        
        if t1 > t2:
            t1, t2 = t2, t1
        
        left, right = self.split(t1)
        
        t2 = (t2 - t1) / (1. - t1)
        
        left, right = right.split(t2)
        
        return left
    
    
    def reduce(self):
        """
        Splits current curve into multiple simple segments, where the simpleness
        is defined as having all control points on the same side of the
        baseline, the control-to-end-point lines may not cross and the angle
        between the end point normals is no greater than 60 degrees.
        
        Returns:
            (pero.Bezier,)
                Collection of simple segments.
        """
        
        if self.simple():
            return tuple((self, ))
        
        pass1 = []
        pass2 = []
        step = REDUCE_STEP
        
        extremes = self.extremes()
        extremes = list(set([0] + extremes[0] + extremes[1] + [1]))
        extremes.sort()
        
        t1 = extremes[0]
        for i in range(1, len(extremes)):
            t2 = extremes[i]
            
            segment = self.slice(t1, t2)
            segment._t1 = t1
            segment._t2 = t2
            
            pass1.append(segment)
            t1 = t2
        
        for p1 in pass1:
            
            if p1.simple():
                pass2.append(p1)
                continue
            
            t1 = 0
            t2 = 0
            while t2 <= 1:
                
                t2 = t1 + step
                while t2 <= 1+step:
                    
                    # not sure about this :)
                    if abs(t1 - t2) < step:
                        t1 = t2
                        break
                    
                    segment = p1.slice(t1, t2)
                    
                    if not segment.simple():
                        t2 -= step
                        
                        if abs(t1 - t2) < step:
                            t1 = t2
                            break
                        
                        segment = p1.slice(t1, t2)
                        segment._t1 = self._scale_t(t1, 0, 1, p1._t1, p1._t2)
                        segment._t2 = self._scale_t(t2, 0, 1, p1._t1, p1._t2)
                        
                        pass2.append(segment)
                        t1 = t2
                        break
                    
                    t2 += step
            
            if t1 < 1:
                
                tmp, segment = p1.split(t1)
                segment._t1 = self._scale_t(t1, 0, 1, p1._t1, p1._t2)
                segment._t2 = p1._t2
                
                pass2.append(segment)
        
        return tuple(pass2)
    
    
    def project(self, x, y):
        """
        Finds the point on current curve, nearest to specified coordinates. This
        gives back the on-curve coordinates, t-value and distance.
        
        Args:
            x: int or float
                X-coordinate of the point.
            
            y: int or float
                Y-coordinate of the point.
        
        Returns:
            (float, float, float, float)
                Results as x,y coordinates, t-value and distance.
        """
        
        lut = self.lut()
        nearest = self._get_nearest_lut(lut, x, y)
        idx = nearest[0]
        dist = nearest[1]
        ln = len(lut)-1.
        
        if idx == 0 or idx == ln:
            t = idx / ln
            p = self.point(t)
            
            return p[0], p[1], t, dist
        
        t1 = (idx - 1.) / ln
        t2 = (idx + 1.) / ln
        t = t1
        ft = t
        step = PROJECTION_STEP/ln
        dist += 1
        
        while t < t2+step:
            
            p = self.point(t)
            dx = x - p[0]
            dy = y - p[1]
            d = numpy.sqrt(dx*dx + dy*dy)
            
            if d < dist:
                dist = d
                ft = t
            
            t += step
        
        p = self.point(ft)
        
        return p[0], p[1], t, dist
    
    
    def overlaps(self, curve):
        """
        Checks if there is any overlap between bounding boxes of current curve
        and given curve.
        
        Args:
            curve: pero.Bezier
                Curve to check.
        
        Returns:
            bool
                Returns True if any overlap of bounding boxes exists, False
                otherwise.
        """
        
        return self.bbox().overlaps(curve.bbox())
    
    
    def cuts(self, x1, y1, x2, y2):
        """
        Finds the intersections between current curve and specified infinite
        line. Intersections are returned as t-values of current curve.
        
        Args:
            x1: int or float
                X-coordinate of the line start point.
            
            y1: int or float
                Y-coordinate of the line start point.
            
            x2: int or float
                X-coordinate of the line end point.
            
            y2: int or float
                Y-coordinate of the line end point.
        
        Returns:
            (float,)
                Intersections as t-values for current curve.
        """
        
        return tuple(helpers.roots((x1, y1), (x2, y2), *self.points))
    
    
    def xcuts(self, x):
        """
        Finds the intersections between current curve and an infinite line going
        through given x-coordinate. Intersections are returned as t-values of
        current curve.
        
        Args:
            x: int, float
                X-coordinate of the cut.
        
        Returns:
            (float,)
                Intersections as t-values for current curve.
        """
        
        return self.cuts(x, -1, x, 1)
    
    
    def ycuts(self, y):
        """
        Finds the intersections between current curve and an infinite line going
        through given y-coordinate. Intersections are returned as t-values of
        current curve.
        
        Args:
            y: int, float
                Y-coordinate of the cut.
        
        Returns:
            (float,)
                Intersections as t-values for current curve.
        """
        
        return self.cuts(-1, y, 1, y)
    
    
    def intersects(self, curve=None, threshold=INTERSECT_THRESHOLD):
        """
        Finds the intersections between current curve and another. Intersections
        are returned as pairs of t-values, where the first corresponds to this
        curve and the second corresponds to the other curve.
        
        If the curve is not specified, self intersections of current curve are
        returned.
        
        Args:
            curve: pero.Bezier or None
                Curve to intersect with current curve.
            
            threshold: float
                Resolution threshold as a maximum size of a segment.
        
        Returns:
            ((float,float),)
                Intersections as t-values for current and given curves.
        """
        
        if curve is None:
            inters = self._intersects_self(threshold)
        else:
            inters = self._intersects_curve(curve, threshold)
        
        if not inters:
            return tuple()
        
        inters.sort()
        buff = [inters[0]]
        t1 = buff[-1]
        
        for t2 in inters[1:]:
            
            if not helpers.equals(t1[0], t2[0], 1e-3):
                buff.append(t2)
                t1 = t2
            
            elif not helpers.equals(t1[1], t2[1], 1e-3):
                buff.append(t2)
                t1 = t2
        
        return tuple(buff)
    
    
    def clone(self):
        """
        Creates exact clone of current curve.
        
        Returns:
            pero.Bezier
                Cloned curve.
        """
        
        curve = Bezier(
            self._x1, self._y1,
            self._cx1, self._cy1,
            self._cx2, self._cy2,
            self._x2, self._y2)
        
        curve._t1 = self._t1
        curve._t2 = self._t2
        
        return curve
    
    
    def _scale_t(self, t, ds, de, ts, te):
        """Scales given t-value into new range."""
    
        return ts + (te-ts)*(t-ds)/(de-ds)
    
    
    def _get_nearest_lut(self, lut, x, y):
        """Gets closest item and distance from look-up table."""
        
        dist = numpy.power(2., 63)
        idx = 0
        
        for i, p in enumerate(lut):
            
            dx = x - p[0]
            dy = y - p[1]
            d = numpy.sqrt(dx*dx + dy*dy)
            
            if d < dist:
                dist = d
                idx = i
        
        return idx, dist
    
    
    def _intersects_self(self, threshold):
        """Calculates all intersection points with self."""
        
        reduced = self.reduce()
        
        buff = []
        for i in range(len(reduced)-2):
            
            seg1 = reduced[i:i+1]
            seg2 = reduced[i+2:]
            
            pairs = []
            for s1 in seg1:
                for s2 in seg2:
                    if s1.overlaps(s2):
                        pairs.append((s1, s2))
            
            for pair in pairs:
                buff += self._intersect_simple(pair[0], pair[1], threshold)
        
        return buff
    
    
    def _intersects_curve(self, curve, threshold):
        """Calculates all intersection points with given curve."""
        
        seg1 = self.reduce()
        seg2 = curve.reduce()
        
        pairs = []
        for s1 in seg1:
            for s2 in seg2:
                if s1.overlaps(s2):
                    pairs.append((s1, s2))
        
        buff = []
        for pair in pairs:
            buff += self._intersect_simple(pair[0], pair[1], threshold)
        
        return buff
    
    
    def _intersect_simple(self, c1, c2, threshold):
        """Calculates all intersection points between two simple segments."""
        
        c1b = c1.bbox()
        c2b = c2.bbox()
        
        if c1b.w + c1b.h < threshold and c2b.w + c2b.h < threshold:
            t1 = 0.5 * (c1._t1 + c1._t2)
            t2 = 0.5 * (c2._t1 + c2._t2)
            return tuple(((t1, t2),))
        
        seg1 = (c1,)
        if c1b.w + c1b.h >= threshold:
            seg1 = c1.split(0.5)
        
        seg2 = (c2,)
        if c2b.w + c2b.h >= threshold:
            seg2 = c2.split(0.5)
        
        pairs = []
        for s1 in seg1:
            for s2 in seg2:
                if s1.overlaps(s2):
                    pairs.append((s1,s2))
        
        buff = []
        for pair in pairs:
            buff += self._intersect_simple(pair[0], pair[1], threshold)
        
        return buff



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

from ... enums import *


SVG_LINE_CAP = {
    LINE_CAP_BUTT: 'butt',
    LINE_CAP_SQUARE: 'square',
    LINE_CAP_ROUND: 'round'}

SVG_LINE_JOIN = {
    LINE_JOIN_BEVEL: 'bevel',
    LINE_JOIN_MITER: 'miter',
    LINE_JOIN_ROUND: 'round'}

SVG_LINE_STYLE = {
    LINE_STYLE_CUSTOM: 'custom',
    LINE_STYLE_SOLID: 'solid',
    LINE_STYLE_DOTTED: DASH_VALUES_DOTTED,
    LINE_STYLE_DASHED: DASH_VALUES_DASHED,
    LINE_STYLE_DASHDOTTED: DASH_VALUES_DASHDOTTED}

SVG_FILL_RULE = {
    FILL_RULE_EVENODD: 'evenodd',
    FILL_RULE_WINDING: 'nonzero'}

SVG_FONT_STYLE = {
    FONT_STYLE_NORMAL: 'normal',
    FONT_STYLE_ITALIC: 'italic'}

SVG_FONT_FAMILY = {
    FONT_FAMILY_SERIF: 'serif',
    FONT_FAMILY_SANS: 'sans-serif',
    FONT_FAMILY_MONO: 'monospace'}

SVG_FONT_WEIGHT = {
    FONT_WEIGHT_NORMAL: 'normal',
    FONT_WEIGHT_LIGHT: 'lighter',
    FONT_WEIGHT_BOLD: 'bold',
    FONT_WEIGHT_BLACK: 'bold',
    FONT_WEIGHT_HEAVY: 'bold',
    FONT_WEIGHT_SEMIBOLD: 'bold',
    FONT_WEIGHT_MEDIUM: 'normal',
    FONT_WEIGHT_ULTRALIGHT: 'lighter',
    FONT_WEIGHT_THIN: 'lighter'}

SVG_TEXT_ALIGN = {
    TEXT_ALIGN_LEFT: 'start',
    TEXT_ALIGN_RIGHT: 'end',
    TEXT_ALIGN_CENTER: 'middle'}

SVG_TEXT_BASE = {
    TEXT_BASE_TOP: 'text-before-edge',
    TEXT_BASE_MIDDLE: 'central',
    TEXT_BASE_BOTTOM: 'text-after-edge'}



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import decimal
from .. enums import *
from .. properties import *
from . utils import *
from . formatter import Formatter


class TimeFormatter(Formatter):
    """
    This formatter tool formats given time in seconds into reasonable time
    representation. The formatting can be fully automatic based on current
    'domain' and 'precision' or it can be controlled by defining the 'template'
    property.
    
    The 'template' should be specified by the parts or granularity to use,
    optionally with additional formatting (e.g. '{h}:{m:.02.0f}:{s:.02.0f}').
    The available parts, which can be used are defined by the pero.TIME
    enum. If there is no specific formatting of a part defined directly by the
    template, a default formatting or a custom sub-template is used. The parts
    sub-templates are definable by corresponding properties (e.g. 'h_template'
    for hours or 's_template' for seconds, s_template='{:.0f}s').
    
    For different purposes, different style of rounding might be necessary. For
    example the normal half-rounding should be used for axis labels, however, in
    case of time counters the value should always be rounded down. This behavior
    can be specified by the 'rounding' property as any item from the
    pero.ROUNDING enum.
    
    Properties:
        
        template: str, None or UNDEF
            Specifies the main template to be used instead of automatic
            formatting.
        
        d_template: str, None or UNDEF
            Specifies the template to be used for days formatting.
        
        h_template: str, None or UNDEF
            Specifies the template to be used for hours formatting.
        
        m_template: str, None or UNDEF
            Specifies the template to be used for minutes formatting.
        
        s_template: str, None or UNDEF
            Specifies the template to be used for seconds formatting.
        
        ms_template: str, None or UNDEF
            Specifies the template to be used for milliseconds formatting.
        
        us_template: str, None or UNDEF
            Specifies the template to be used for microseconds formatting.
        
        ns_template: str, None or UNDEF
            Specifies the template to be used for nanoseconds formatting.
        
        rounding: str
            Specifies the rounding style as any item from the pero.ROUNDING
            enum.
        
        separator: str or UNDEF
            Specifies the separator to be used between individual parts.
        
        add_units: bool
            Specifies whether the units should be added to individual parts.
    """
    
    template = StringProperty(UNDEF, dynamic=False, nullable=True)
    d_template = StringProperty(UNDEF, dynamic=False, nullable=True)
    h_template = StringProperty(UNDEF, dynamic=False, nullable=True)
    m_template = StringProperty(UNDEF, dynamic=False, nullable=True)
    s_template = StringProperty(UNDEF, dynamic=False, nullable=True)
    ms_template = StringProperty(UNDEF, dynamic=False, nullable=True)
    us_template = StringProperty(UNDEF, dynamic=False, nullable=True)
    ns_template = StringProperty(UNDEF, dynamic=False, nullable=True)
    
    rounding = EnumProperty(ROUNDING.HALFUP, enum=ROUNDING, dynamic=False)
    separator = StringProperty(UNDEF, dynamic=False)
    add_units = BoolProperty(False, dynamic=False)
    
    
    def __init__(self, **overrides):
        """Initializes a new instance of TimeFormatter."""
        
        super().__init__(**overrides)
        
        # init buffers
        self._template = None
        self._templates_multi = None
        self._templates_single = None
        
        self._is_dirty = True
        
        # get available parts (units, factor)
        self._parts = sorted(SPLITS_TIME.items(), key=lambda d: d[1], reverse=True)
        
        # bind events
        self.bind(EVT_PROPERTY_CHANGED, self._on_time_formatter_property_changed)
    
    
    def format(self, value, *args, **kwargs):
        """
        Formats a given value using time formatting.
        
        Args:
            value: float
                Time in seconds.
        
        Returns:
            str
                Formatted label.
        """
        
        # init formatting
        if self._is_dirty:
            self._init_formatting()
        
        # get template
        template = self._template
        if not template:
            template = self._make_template(value)
        
        # init parts
        parts = {x[0]: 0 for x in self._parts}
        
        # split time
        last = None
        for units, f in self._parts:
            
            key = "{%s:" % units
            if key in template:
                
                # get part value
                val = value / float(f)
                value -= int(val) * f
                parts[units] = decimal.Decimal(val)
                
                # remove fractions from higher part
                if last:
                    parts[last] = int(parts[last])
                
                last = units
        
        # init context
        context = decimal.Context()
        
        if self.rounding == ROUNDING.FLOOR:
            context.rounding = decimal.ROUND_DOWN
        elif self.rounding == ROUNDING.CEIL:
            context.rounding = decimal.ROUND_UP
        
        # format with correct rounding
        with decimal.localcontext(context):
            return template.format(**parts)
    
    
    def _init_formatting(self):
        """Initializes formatting based on current settings."""
        
        # reset
        self._template = None
        self._is_dirty = False
        
        # init templates
        self._init_templates()
        
        # use custom template
        if self.template:
            self._template = self._expand_template(self.template, False)
            return
        
        # check domain
        if not self.domain:
            return
        
        # make template
        self._template = self._make_template(abs(self.domain))
    
    
    def _init_templates(self):
        """Initializes templates."""
        
        # init default parts templates
        self._templates_multi = {
            DAYS: "{%s:.0f}" % DAYS,
            HOURS: "{%s:.0f}" % HOURS,
            MINUTES: "{%s:02.0f}" % MINUTES,
            SECONDS: "{%s:02.0f}" % SECONDS,
            MSECONDS: "{%s:03.0f}" % MSECONDS,
            USECONDS: "{%s:03.0f}" % USECONDS,
            NSECONDS: "{%s:03.0f}" % NSECONDS}
        
        # init default singles templates
        self._templates_single = {x[0]: ("{%s:.2f}" % x[0]) for x in self._parts}
        
        # get user-defined templates
        for units, f in self._parts:
            
            # get part template
            template = self.get_property(units+"_template")
            if not template:
                continue
            
            # ensure part name is present
            template = template.replace("{:", "{%s:" % units)
            
            # store template
            self._templates_multi[units] = template
            self._templates_single[units] = template
    
    
    def _make_template(self, domain):
        """Creates template to cover expected range."""
        
        # check current precision
        precision = domain
        if self.precision and self.precision < domain:
            precision = self.precision
        
        # get required parts
        template = []
        for units, f in self._parts:
            
            if domain >= f:
                template.append(units)
            
            if precision >= f:
                break
        
        # get separator
        separator = self.separator
        if separator is UNDEF:
            separator = " " if self.add_units else ":"
        
        # init template
        template = separator.join("{%s}" % x for x in template)
        
        # expand parts
        return self._expand_template(template, self.add_units)
    
    
    def _expand_template(self, template, add_units):
        """Expands template parts."""
        
        # replace parts in template
        for key, tmpl in self._templates_multi.items():
            
            # make full key
            tag = "{%s}" % key
            
            # add units
            if add_units:
                tmpl = tmpl + " %s" % key
            
            # check for singles
            if template == tag:
                units = " %s" % key if add_units else ""
                tmpl = self._templates_single[key] + units
                return template.replace(tag, tmpl)
            
            # replace in template
            template = template.replace(tag, tmpl)
        
        return template
    
    
    def _on_time_formatter_property_changed(self, evt=None):
        """Called after a property has changed."""
        
        self._is_dirty = True



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import unittest
import pero


class TestCase(unittest.TestCase):
    """Test case for ticker helpers."""
    
    
    def test_step_size(self):
        """Tests whether step size calculation works correctly."""
        
        # test positive
        step = pero.tickers.calc_step_size(10, 1, (5,3,2,1))
        self.assertEqual(step, 10)
        
        step = pero.tickers.calc_step_size(10, 2, (5,3,2,1))
        self.assertEqual(step, 5)
        
        step = pero.tickers.calc_step_size(10, 3, (5,3,2,1))
        self.assertEqual(step, 3)
        
        step = pero.tickers.calc_step_size(10, 4, (5,3,2,1))
        self.assertEqual(step, 3)
        
        step = pero.tickers.calc_step_size(10, 5, (5,3,2,1))
        self.assertEqual(step, 2)
        
        step = pero.tickers.calc_step_size(10, 6, (5,3,2,1))
        self.assertEqual(step, 2)
        
        step = pero.tickers.calc_step_size(10, 7, (5,3,2,1))
        self.assertEqual(step, 2)
        
        step = pero.tickers.calc_step_size(10, 8, (5,3,2,1))
        self.assertEqual(step, 1)
        
        step = pero.tickers.calc_step_size(10, 9, (5,3,2,1))
        self.assertEqual(step, 1)
        
        step = pero.tickers.calc_step_size(10, 10, (5,3,2,1))
        self.assertEqual(step, 1)
    
    
    def test_lin_ticks(self):
        """Tests whether linear ticks are calculated correctly."""
        
        # test basic
        ticks = pero.tickers.make_lin_ticks(0, 5, 1)
        model = (0, 1, 2, 3, 4, 5)
        self.assertEqual(ticks, model)
        
        # test start
        ticks = pero.tickers.make_lin_ticks(0.5, 5, 1)
        model = (1, 2, 3, 4, 5)
        self.assertEqual(ticks, model)
        
        # test end
        ticks = pero.tickers.make_lin_ticks(0, 5.5, 1)
        model = (0, 1, 2, 3, 4, 5)
        self.assertEqual(ticks, model)
        
        # test reversed range
        ticks = pero.tickers.make_lin_ticks(5.5, 0.5, 1)
        model = (5, 4, 3, 2, 1)
        self.assertEqual(ticks, model)


# run test case
if __name__ == "__main__":
    unittest.main(verbosity=2)



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

from . enum import Enum

# define events
EVT_PROPERTY_CHANGED = 'evt_prop_changed'
EVT_PEN_CHANGED = 'evt_pen_changed'
EVT_BRUSH_CHANGED = 'evt_brush_changed'
EVT_TEXT_CHANGED = 'evt_text_changed'
EVT_VIEW = 'evt_view'
EVT_SIZE = 'evt_size'
EVT_ZOOM = 'evt_zoom'
EVT_KEY = 'evt_key'
EVT_KEY_DOWN = 'evt_key_down'
EVT_KEY_UP = 'evt_key_up'
EVT_MOUSE = 'evt_mouse'
EVT_MOUSE_MOTION = 'evt_mouse_motion'
EVT_MOUSE_SCROLL = 'evt_mouse_scroll'
EVT_MOUSE_ENTER = 'evt_mouse_enter'
EVT_MOUSE_LEAVE = 'evt_mouse_leave'
EVT_LEFT_DOWN = 'evt_left_down'
EVT_LEFT_UP = 'evt_left_up'
EVT_LEFT_DCLICK = 'evt_left_dclick'
EVT_MIDDLE_DOWN = 'evt_middle_down'
EVT_MIDDLE_UP = 'evt_middle_up'
EVT_MIDDLE_DCLICK = 'evt_middle_dclick'
EVT_RIGHT_DOWN = 'evt_right_down'
EVT_RIGHT_UP = 'evt_right_up'
EVT_RIGHT_DCLICK = 'evt_right_dclick'
EVT_TOUCH = 'evt_touch'
EVT_TOUCH_START = 'evt_touch_start'
EVT_TOUCH_END = 'evt_touch_end'
EVT_TOUCH_MOVE = 'evt_touch_move'
EVT_TOUCH_CANCEL = 'evt_touch_cancel'
EVT_TOUCH_DTAP = 'evt_touch_dtap'

EVENT = Enum(
    PROPERTY_CHANGED = EVT_PROPERTY_CHANGED,
    PEN_CHANGED = EVT_PEN_CHANGED,
    BRUSH_CHANGED = EVT_BRUSH_CHANGED,
    TEXT_CHANGED = EVT_TEXT_CHANGED,
    VIEW = EVT_VIEW,
    SIZE = EVT_SIZE,
    ZOOM = EVT_ZOOM,
    KEY = EVT_KEY,
    KEY_DOWN = EVT_KEY_DOWN,
    KEY_UP = EVT_KEY_UP,
    MOUSE = EVT_MOUSE,
    MOUSE_MOTION = EVT_MOUSE_MOTION,
    MOUSE_SCROLL = EVT_MOUSE_SCROLL,
    MOUSE_ENTER = EVT_MOUSE_ENTER,
    MOUSE_LEAVE = EVT_MOUSE_LEAVE,
    LEFT_DOWN = EVT_LEFT_DOWN,
    LEFT_UP = EVT_LEFT_UP,
    LEFT_DCLICK = EVT_LEFT_DCLICK,
    MIDDLE_DOWN = EVT_MIDDLE_DOWN,
    MIDDLE_UP = EVT_MIDDLE_UP,
    MIDDLE_DCLICK = EVT_MIDDLE_DCLICK,
    RIGHT_DOWN = EVT_RIGHT_DOWN,
    RIGHT_UP = EVT_RIGHT_UP,
    RIGHT_DCLICK = EVT_RIGHT_DCLICK,
    TOUCH = EVT_TOUCH,
    TOUCH_START = EVT_TOUCH_START,
    TOUCH_END = EVT_TOUCH_END,
    TOUCH_MOVE = EVT_TOUCH_MOVE,
    TOUCH_CANCEL = EVT_TOUCH_CANCEL,
    TOUCH_DTAP = EVT_TOUCH_DTAP)

# define touch states
TOUCH_PRESSED = 'pressed'
TOUCH_RELEASED = 'released'
TOUCH_MOVED = 'moved'
TOUCH_STATIC = 'static'

TOUCH = Enum(
    PRESSED = TOUCH_PRESSED,
    RELEASED = TOUCH_RELEASED,
    MOVED = TOUCH_MOVED,
    STATIC = TOUCH_STATIC)



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import numpy
from .. enums import *
from .. properties import *
from . glyph import Glyph


class Axis(Glyph):
    """
    Abstract base class for various types of axes. They serve as a drawing glyph
    only and do not implement any logic to generate ticks and labels. Everything
    is expected to be supplied in its final form already.
    
    Properties:
        
        show_line: bool or callable
            Specifies whether the major axis line should be displayed.
        
        show_labels: bool or callable
            Specifies whether the labels should be displayed.
        
        show_major_ticks: bool or callable
            Specifies whether the major ticks should be displayed.
        
        show_minor_ticks: bool or callable
            Specifies whether the minor ticks should be displayed.
        
        x: int, float or callable
            Specifies the x-coordinate of the origin.
        
        y: int, float or callable
            Specifies the y-coordinate of the origin.
        
        line properties:
            Includes pero.LineProperties to specify the main line.
        
        labels: (str,), callable, None or UNDEF
            Specifies the major ticks labels.
        
        label_offset: int, float or callable
            Specifies the shift of the labels from the main line.
        
        label_flip: bool or callable
            Specifies whether the labels should be displayed on default side of
            the axis according to other properties (False) or flipped to the
            other side (True).
        
        label_text properties:
            Includes pero.TextProperties to specify the labels text
            properties. Some of them (e.g. alignment, angle, baseline) are
            typically set automatically according to other axis properties.
        
        major_ticks: (float,), callable, None or UNDEF
            Specifies the coordinates of the major ticks.
        
        major_tick_size: int, float or callable
            Specifies the length of the major ticks.
        
        major_tick_offset: int, float or callable
            Specifies the shift of the major ticks from the main line.
        
        major_tick_flip: bool or callable
            Specifies whether the major ticks should be displayed on default
            side of the axis according to other properties (False) or flipped to
            the other side (True).
        
        major_tick_line properties:
            Includes pero.LineProperties to specify the major ticks line.
        
        minor_ticks: (float,), callable, None or UNDEF
            Specifies the coordinates of the minor ticks.
        
        minor_tick_size: int, float or callable
            Specifies the length of the minor ticks.
        
        minor_tick_offset: int, float or callable
            Specifies the shift of the minor ticks from the main line.
        
        minor_tick_flip: bool or callable
            Specifies whether the minor ticks should be displayed on default
            side of the axis according to other properties (False) or flipped to
            the other side (True).
        
        minor_tick_line properties:
            Includes pero.LineProperties to specify the minor ticks line.
    """
    
    show_line = BoolProperty(True)
    show_labels = BoolProperty(True)
    show_major_ticks = BoolProperty(True)
    show_minor_ticks = BoolProperty(True)
    
    x = NumProperty(0)
    y = NumProperty(0)
    
    line = Include(LineProperties, line_color="#000")
    
    labels = TupleProperty(None, nullable=True, intypes=(str,))
    label_text = Include(TextProperties, prefix="label_", font_size=11, text_align=UNDEF, text_base=UNDEF)
    label_offset = NumProperty(10)
    label_flip = BoolProperty(False)
    
    major_ticks = TupleProperty(None, nullable=True, intypes=(int, float))
    major_tick_line = Include(LineProperties, prefix="major_tick_", line_color="#000")
    major_tick_size = NumProperty(5)
    major_tick_offset = NumProperty(0)
    major_tick_flip = BoolProperty(False)
    
    minor_ticks = TupleProperty(None, nullable=True, intypes=(int, float))
    minor_tick_line = Include(LineProperties, prefix="minor_tick_", line_color="#000")
    minor_tick_size = NumProperty(3)
    minor_tick_offset = NumProperty(0)
    minor_tick_flip = BoolProperty(False)


class StraitAxis(Axis):
    """
    Strait axis is a standard type of axis used for Cartesian plots. By default
    the axis is drawn as a horizontal line with ticks and labels facing down. To
    create a vertical or angled axis either the 'position' or 'angle' property
    must be specified and the whole axis is rotated around the origin given by
    'x' and 'y' properties.
    
    According to the 'relative' property, the axis ticks are expected to be
    provided as relative values (True) (as a distance from axis origin) or as
    absolute values (False) (as a distance from device zero).
    
    Properties:
        
        show_title: bool or callable
            Specifies whether the title should be displayed.
        
        angle properties:
            Includes pero.AngleProperties to specify the angle.
        
        length: int, float or callable
            Specifies the main axis line length.
        
        offset: int, float or callable
            Specifies the main axis line shift from the origin. (This shift is
            not applied to the ticks.)
        
        position: pero.POSITION_LRTB or callable
            Specifies the axis base orientation and labels and titles position
            as any item from the pero.POSITION_LRTB enum.
        
        relative: bool or callable
            Specifier whether the ticks values are given as a shift from the
            axis origin (True) or as absolute values (False).
        
        label_overlap: bool or callable
            Specifies whether the labels can overlap (True) each other or should
            be removed automatically (False).
        
        label_angle properties:
            Includes pero.AngleProperties to specify the labels angle.
        
        title: str, callable, None or UNDEF
            Specifies the title to show.
        
        title_position: pero.POSITION_SEM or callable
            Specifies the title position relative to the axis origin as any item
            from the pero.POSITION_SEM enum.
        
        title_offset: int, float or callable
            Specifies the shift of the title from the main axis line.
        
        title_flip: bool or callable
            Specifies whether the title should be displayed on default side
            of the axis according to other properties (False) or flipped to the
            other side (True).
        
        title_text properties:
            Includes pero.TextProperties to specify the title text
            properties. Some of them (e.g. alignment, angle, baseline) are
            typically set automatically according to other axis properties.
        
        title_angle properties:
            Includes pero.AngleProperties to specify the title angle. By
            default, the angle is set automatically according to other axis
            properties.
    """
    
    show_title = BoolProperty(True)
    
    position = EnumProperty(POS_BOTTOM, enum=POSITION_LRTB)
    relative = BoolProperty(False)
    
    angle = Include(AngleProperties)
    length = NumProperty(UNDEF)
    offset = NumProperty(0)
    
    title = StringProperty(None, nullable=True)
    title_position = EnumProperty(POS_MIDDLE, enum=POSITION_SEM)
    title_offset = NumProperty(25)
    title_flip = BoolProperty(False)
    title_text = Include(TextProperties, prefix="title_", font_weight=FONT_WEIGHT_BOLD, font_size=12, text_align=UNDEF, text_base=UNDEF)
    title_angle = Include(AngleProperties, prefix="title_", angle=UNDEF)
    
    label_angle = Include(AngleProperties, prefix="label_")
    label_overlap = BoolProperty(False)
    
    
    def draw(self, canvas, source=UNDEF, **overrides):
        """Uses given canvas to draw the axis."""
        
        # check if visible
        if not self.is_visible(source, overrides):
            return
        
        # get properties
        tag = self.get_property('tag', source, overrides)
        show_line = self.get_property('show_line', source, overrides)
        show_title = self.get_property('show_title', source, overrides)
        show_labels = self.get_property('show_labels', source, overrides)
        show_major_ticks = self.get_property('show_major_ticks', source, overrides)
        show_minor_ticks = self.get_property('show_minor_ticks', source, overrides)
        
        # start drawing group
        canvas.group(tag, "axis")
        
        # draw minor ticks
        if show_minor_ticks:
            canvas.group(None, "minor_ticks")
            self._draw_minor_ticks(canvas, source, overrides)
            canvas.ungroup()
        
        # draw major ticks
        if show_major_ticks:
            canvas.group(None, "major_ticks")
            self._draw_major_ticks(canvas, source, overrides)
            canvas.ungroup()
        
        # draw labels
        if show_labels:
            canvas.group(None, "labels")
            self._draw_labels(canvas, source, overrides)
            canvas.ungroup()
        
        # draw title
        if show_title:
            self._draw_title(canvas, source, overrides)
        
        # draw main line
        if show_line:
            self._draw_line(canvas, source, overrides)
        
        # end drawing group
        canvas.ungroup()
    
    
    def _draw_major_ticks(self, canvas, source, overrides):
        """Draws axis major ticks."""
        
        # get properties
        ticks = self.get_property('major_ticks', source, overrides)
        size = self.get_property('major_tick_size', source, overrides)
        offset = self.get_property('major_tick_offset', source, overrides)
        flip = self.get_property('major_tick_flip', source, overrides)
        
        # set pen
        canvas.set_pen_by(self, prefix='major_tick_', source=source, overrides=overrides)
        
        # draw ticks
        self._draw_ticks(canvas, source, overrides, ticks, size, offset, flip)
    
    
    def _draw_minor_ticks(self, canvas, source, overrides):
        """Draws axis minor ticks."""
        
        # get properties
        flip = self.get_property('minor_tick_flip', source, overrides)
        ticks = self.get_property('minor_ticks', source, overrides)
        size = self.get_property('minor_tick_size', source, overrides)
        offset = self.get_property('minor_tick_offset', source, overrides)
        
        # set pen
        canvas.set_pen_by(self, prefix='minor_tick_', source=source, overrides=overrides)
        
        # draw ticks
        self._draw_ticks(canvas, source, overrides, ticks, size, offset, flip)
    
    
    def _draw_ticks(self, canvas, source, overrides, ticks, tick_size, tick_offset, tick_flip):
        """Draws axis ticks."""
        
        # check data
        if not ticks:
            return
        
        # get properties
        x = self.get_property('x', source, overrides)
        y = self.get_property('y', source, overrides)
        position = self.get_property('position', source, overrides)
        relative = self.get_property('relative', source, overrides)
        angle = AngleProperties.get_angle(self, '', ANGLE_RAD, source, overrides)
        
        # apply position
        if position in POSITION_LR:
            angle += 0.5*math.pi
        
        if position in POSITION_TR:
            tick_size *= -1
            tick_offset *= -1
        
        # apply flipping
        if tick_flip:
            tick_size *= -1
            tick_offset *= -1
        
        # make ticks relative
        offset = 0
        if not relative:
            offset -= y if position in POSITION_LR else x
        
        # calc sin and cos
        sin = round(math.sin(angle), 5)
        cos = round(math.cos(angle), 5)
        
        # draw ticks
        for tick in ticks:
            
            x1 = x + (tick+offset) * cos - tick_offset * sin
            y1 = y + (tick+offset) * sin + tick_offset * cos
            x2 = x + (tick+offset) * cos - (tick_offset+tick_size) * sin
            y2 = y + (tick+offset) * sin + (tick_offset+tick_size) * cos
            
            canvas.draw_line(x1=x1, y1=y1, x2=x2, y2=y2)
    
    
    def _draw_labels(self, canvas, source, overrides):
        """Draws axis labels."""
        
        # get properties
        x = self.get_property('x', source, overrides)
        y = self.get_property('y', source, overrides)
        position = self.get_property('position', source, overrides)
        relative = self.get_property('relative', source, overrides)
        angle = AngleProperties.get_angle(self, '', ANGLE_RAD, source, overrides)
        
        labels = self.get_property('labels', source, overrides)
        label_offset = self.get_property('label_offset', source, overrides)
        label_angle = AngleProperties.get_angle(self, 'label_', ANGLE_RAD, source, overrides)
        label_flip = self.get_property('label_flip', source, overrides)
        label_overlap = self.get_property('label_overlap', source, overrides)
        label_text_align = self.get_property('label_text_align', source, overrides)
        label_text_base = self.get_property('label_text_base', source, overrides)
        ticks = self.get_property('major_ticks', source, overrides)
        
        # check data
        if not labels:
            return
        
        # set font
        canvas.set_text_by(self, prefix="label_", source=source, overrides=overrides)
        
        # apply position
        if position in POSITION_LR:
            angle += 0.5*math.pi
        
        if position in POSITION_TR:
            label_offset *= -1
        
        # apply flipping
        if label_flip:
            label_offset *= -1
        
        # make ticks relative
        offset = 0
        if not relative:
            offset -= y if position in POSITION_LR else x
        
        # calc sin and cos
        sin = round(math.sin(angle), 5)
        cos = round(math.cos(angle), 5)
        
        # get final orientation
        is_horizontal = sin == 0
        is_vertical = cos == 0
        is_left = cos < 0
        is_bottom = sin > 0
        is_flipped = bool(label_flip) != bool(position in POSITION_TR)
        
        # set alignment
        if label_text_align is UNDEF:
            
            if is_horizontal:
                canvas.text_align = TEXT_ALIGN_CENTER
            elif is_bottom:
                canvas.text_align = TEXT_ALIGN_LEFT if is_flipped else TEXT_ALIGN_RIGHT
            else:
                canvas.text_align = TEXT_ALIGN_RIGHT if is_flipped else TEXT_ALIGN_LEFT
        
        # set baseline
        if label_text_base is UNDEF:
            
            if is_vertical:
                canvas.text_base = TEXT_BASE_MIDDLE
            elif is_left:
                canvas.text_base = TEXT_BASE_TOP if is_flipped else TEXT_BASE_BOTTOM
            else:
                canvas.text_base = TEXT_BASE_BOTTOM if is_flipped else TEXT_BASE_TOP
        
        # draw labels
        area = []
        for i in range(min(len(labels), len(ticks))):
            
            # get values
            label = labels[i]
            pos = ticks[i]
            
            # check label
            if not label:
                continue
            
            # calc anchor
            x1 = x + (pos+offset) * cos - label_offset * sin
            y1 = y + (pos+offset) * sin + label_offset * cos
            
            # get bbox
            bbox = canvas.get_text_bbox(label, x1, y1, label_angle)
            
            # check overlaps
            if not label_overlap and any(bbox.overlaps(box) for box in area):
                continue
            
            # draw label
            canvas.draw_text(label, x=x1, y=y1, angle=label_angle)
            area.append(bbox)
    
    
    def _draw_title(self, canvas, source, overrides):
        """Draws axis title."""
        
        # get properties
        x = self.get_property('x', source, overrides)
        y = self.get_property('y', source, overrides)
        offset = self.get_property('offset', source, overrides)
        position = self.get_property('position', source, overrides)
        length = self.get_property('length', source, overrides)
        angle = AngleProperties.get_angle(self, '', ANGLE_RAD, source, overrides)
        
        title = self.get_property('title', source, overrides)
        title_position = self.get_property('title_position', source, overrides)
        title_offset = self.get_property('title_offset', source, overrides)
        title_angle = AngleProperties.get_angle(self, 'title_', ANGLE_RAD, source, overrides)
        title_flip = self.get_property('title_flip', source, overrides)
        title_text_align = self.get_property('title_text_align', source, overrides)
        title_text_base = self.get_property('title_text_base', source, overrides)
        
        # check data
        if not title:
            return
        
        # set font
        canvas.set_text_by(self, prefix="title_", source=source, overrides=overrides)
        
        # apply position
        if position in POSITION_LR:
            angle += 0.5*math.pi
        
        if position in POSITION_TR:
            title_offset *= -1
        
        # apply flipping
        if title_flip:
            title_offset *= -1
        
        # calc sin and cos
        sin = round(math.sin(angle), 5)
        cos = round(math.cos(angle), 5)
        
        # get final orientation
        is_vertical = cos == 0
        is_left = cos < 0
        is_flipped = bool(title_flip) != bool(position in POSITION_TR)
        
        # get y-offset
        y_offset = title_offset
        
        # get x-offset
        x_offset = offset
        if title_position == POS_MIDDLE:
            x_offset += .5 * length
        elif title_position == POS_END:
            x_offset += length
        
        # calc anchor
        x = x + x_offset * cos - y_offset * sin
        y = y + x_offset * sin + y_offset * cos
        
        # set alignment
        if title_text_align is UNDEF:
            
            if title_position == POS_MIDDLE:
                canvas.text_align = TEXT_ALIGN_CENTER
            
            elif title_position == POS_START and is_vertical:
                canvas.text_align = TEXT_ALIGN_LEFT if is_flipped else TEXT_ALIGN_RIGHT
            
            elif title_position == POS_START:
                canvas.text_align = TEXT_ALIGN_RIGHT if is_left else TEXT_ALIGN_LEFT
            
            elif title_position == POS_END and is_vertical:
                canvas.text_align = TEXT_ALIGN_RIGHT if is_flipped else TEXT_ALIGN_LEFT
            
            elif title_position == POS_END:
                canvas.text_align = TEXT_ALIGN_LEFT if is_left else TEXT_ALIGN_RIGHT
        
        # set baseline
        if title_text_base is UNDEF:
            
            if is_vertical:
                canvas.text_base = TEXT_BASE_BOTTOM
            elif is_left:
                canvas.text_base = TEXT_BASE_TOP if is_flipped else TEXT_BASE_BOTTOM
            else:
                canvas.text_base = TEXT_BASE_BOTTOM if is_flipped else TEXT_BASE_TOP
        
        # get angle
        if title_angle is UNDEF:
            
            if is_vertical:
                title_angle = angle if is_flipped else -angle
            else:
                title_angle = angle-math.pi if is_left else angle
        
        # draw title
        canvas.draw_text(title, x=x, y=y, angle=title_angle)
    
    
    def _draw_line(self, canvas, source, overrides):
        """Draws axis line."""
        
        # get properties
        x = self.get_property('x', source, overrides)
        y = self.get_property('y', source, overrides)
        position = self.get_property('position', source, overrides)
        offset = self.get_property('offset', source, overrides)
        length = self.get_property('length', source, overrides)
        angle = AngleProperties.get_angle(self, '', ANGLE_RAD, source, overrides)
        
        # set pen
        canvas.set_pen_by(self, source=source, overrides=overrides)
        
        # apply position
        if position in POSITION_LR:
            angle -= 0.5*math.pi
            length *= -1
            offset *= -1
        
        # calc sin and cos
        sin = round(math.sin(angle), 5)
        cos = round(math.cos(angle), 5)
        
        # calc coords
        x1 = x + offset * cos
        y1 = y + offset * sin
        x2 = x + (length+offset) * cos
        y2 = y + (length+offset) * sin
        
        # draw line
        canvas.draw_line(x1=x1, y1=y1, x2=x2, y2=y2)


class RadialAxis(Axis):
    """
    Radial axis is a standard type of axis used for polar plots. By default
    the axis is drawn as a circle or arc line with ticks and labels facing out.
    
    The ticks are expected to be provided as absolute angle values in the units
    specified by the 'units' property.
    
    Properties:
        
        radius: int, float or callable
            Specifies the axis radius.
        
        units: str or callable
            Specifies the angle units for the ticks as any item from the
            pero.ANGLE enum.
        
        start_angle properties:
            Includes pero.AngleProperties to specify the start angle.
        
        end_angle properties:
            Includes pero.AngleProperties to specify the end angle.
        
        clockwise: bool or callable
            Specifies the drawing direction. If set to True the axis is drawn
            clockwise, otherwise anti-clockwise.
        
        label_rotation: pero.TEXT_ROTATION or callable
            Specifies the label rotation style as any item from the
            pero.TEXT_ROTATION enum.
    """
    
    radius = NumProperty(100)
    units = EnumProperty(ANGLE_RAD, enum=ANGLE)
    start_angle = Include(AngleProperties, prefix="start_")
    end_angle = Include(AngleProperties, prefix="end_")
    clockwise = BoolProperty(True)
    
    label_rotation = EnumProperty(TEXT_ROT_FOLLOW, enum=TEXT_ROTATION)
    
    
    def draw(self, canvas, source=UNDEF, **overrides):
        """Uses given canvas to draw the axis."""
        
        # check if visible
        if not self.is_visible(source, overrides):
            return
        
        # get properties
        tag = self.get_property('tag', source, overrides)
        show_line = self.get_property('show_line', source, overrides)
        show_labels = self.get_property('show_labels', source, overrides)
        show_major_ticks = self.get_property('show_major_ticks', source, overrides)
        show_minor_ticks = self.get_property('show_minor_ticks', source, overrides)
        
        # start drawing group
        canvas.group(tag, "axis")
        
        # draw minor ticks
        if show_minor_ticks:
            canvas.group(None, "minor_ticks")
            self._draw_minor_ticks(canvas, source, overrides)
            canvas.ungroup()
        
        # draw major ticks
        if show_major_ticks:
            canvas.group(None, "major_ticks")
            self._draw_major_ticks(canvas, source, overrides)
            canvas.ungroup()
        
        # draw labels
        if show_labels:
            canvas.group(None, "labels")
            self._draw_labels(canvas, source, overrides)
            canvas.ungroup()
        
        # draw main line
        if show_line:
            self._draw_line(canvas, source, overrides)
        
        # end drawing group
        canvas.ungroup()
    
    
    def _draw_major_ticks(self, canvas, source, overrides):
        """Draws axis major ticks."""
        
        # get properties
        ticks = self.get_property('major_ticks', source, overrides)
        size = self.get_property('major_tick_size', source, overrides)
        offset = self.get_property('major_tick_offset', source, overrides)
        flip = self.get_property('major_tick_flip', source, overrides)
        
        # set pen
        canvas.set_pen_by(self, prefix='major_tick_', source=source, overrides=overrides)
        
        # draw ticks
        self._draw_ticks(canvas, source, overrides, ticks, size, offset, flip)
    
    
    def _draw_minor_ticks(self, canvas, source, overrides):
        """Draws axis minor ticks."""
        
        # get properties
        ticks = self.get_property('minor_ticks', source, overrides)
        size = self.get_property('minor_tick_size', source, overrides)
        offset = self.get_property('minor_tick_offset', source, overrides)
        flip = self.get_property('minor_tick_flip', source, overrides)
        
        # set pen
        canvas.set_pen_by(self, prefix='minor_tick_', source=source, overrides=overrides)
        
        # draw ticks
        self._draw_ticks(canvas, source, overrides, ticks, size, offset, flip)
    
    
    def _draw_ticks(self, canvas, source, overrides, ticks, tick_size, tick_offset, tick_flip):
        """Draws axis ticks."""
        
        # check data
        if not ticks:
            return
        
        # get properties
        x = self.get_property('x', source, overrides)
        y = self.get_property('y', source, overrides)
        radius = self.get_property('radius', source, overrides)
        units = self.get_property('units', source, overrides)
        
        # convert angles
        if units == ANGLE_DEG:
            ticks = tuple(map(math.radians, ticks))
        
        # get radii
        if tick_flip:
            inner_radius = radius - tick_offset - tick_size
            outer_radius = inner_radius + tick_size
        else:
            inner_radius = radius + tick_offset
            outer_radius = inner_radius + tick_size
        
        # draw ticks
        for angle in ticks:
            
            cos = math.cos(angle)
            sin = math.sin(angle)
            
            x1 = x + inner_radius * cos
            y1 = y + inner_radius * sin
            x2 = x + outer_radius * cos
            y2 = y + outer_radius * sin
            
            canvas.draw_line(x1, y1, x2, y2)
    
    
    def _draw_labels(self, canvas, source, overrides):
        """Draws axis labels."""
        
        # get properties
        x = self.get_property('x', source, overrides)
        y = self.get_property('y', source, overrides)
        radius = self.get_property('radius', source, overrides)
        
        flip = self.get_property('label_flip', source, overrides)
        labels = self.get_property('labels', source, overrides)
        label_offset = self.get_property('label_offset', source, overrides)
        label_rotation = self.get_property('label_rotation', source, overrides)
        ticks = self.get_property('major_ticks', source, overrides)
        units = self.get_property('units', source, overrides)
        
        # check data
        if not labels:
            return
        
        # set font
        canvas.set_text_by(self, prefix="label_", source=source, overrides=overrides)
        
        # convert angles
        if units == ANGLE_DEG:
            ticks = tuple(map(math.radians, ticks))
        
        # get position
        position = POS_INSIDE if flip else POS_OUTSIDE
        
        # get radius
        radius += -label_offset if flip else label_offset
        
        # draw labels
        for i in range(min(len(labels), len(ticks))):
            
            # get values
            label = labels[i]
            angle = ticks[i]
            
            # check label
            if not label:
                continue
            
            # draw label
            canvas.draw_text_polar(label, x, y, radius, angle, position, label_rotation)
    
    
    def _draw_line(self, canvas, source, overrides):
        """Draws axis line."""
        
        # get properties
        x = self.get_property('x', source, overrides)
        y = self.get_property('y', source, overrides)
        radius = self.get_property('radius', source, overrides)
        clockwise = self.get_property('clockwise', source, overrides)
        start_angle = AngleProperties.get_angle(self, 'start_', ANGLE_RAD, source, overrides)
        end_angle = AngleProperties.get_angle(self, 'end_', ANGLE_RAD, source, overrides)
        
        # set pen
        canvas.set_pen_by(self, source=source, overrides=overrides)
        
        # clear fill color
        canvas.fill_color = None
        
        # draw full circle
        if abs(start_angle - end_angle) >= 2*math.pi:
            canvas.draw_circle(x, y, radius)
        
        # draw arc
        else:
            canvas.draw_arc(x, y, radius, start_angle, end_angle, clockwise)
